__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from .ply import yacc

from nml import expression, generic, nmlop, tokens, unit
from nml.actions import actionD, real_sprite
from nml.ast import (
    alt_sprites,
    assignment,
    base_graphics,
    basecost,
    cargotable,
    conditional,
    constant,
    deactivate,
    disable_item,
    error,
    font,
    general,
    grf,
    item,
    loop,
    override,
    produce,
    replace,
    skipall,
    snowline,
    sort_vehicles,
    spriteblock,
    switch,
    tilelayout,
    townnames,
    tracktypetable,
)


class NMLParser:
    """
    @ivar lexer: Scanner providing tokens.
    @type lexer: L{NMLLexer}

    @ivar tokens: Tokens of the scanner (used by PLY).
    @type tokens: C{List} of C{str}

    @ivar parser: PLY parser.
    @type parser: L{ply.yacc}
    """

    def __init__(self, debug=False):
        self.lexer = tokens.NMLLexer()
        self.lexer.build()
        self.tokens = self.lexer.tokens
        self.parser = yacc.yacc(
            module=self,
            debug=debug,
            optimize=not debug,
        )

    def parse(self, text, input_filename):
        self.lexer.setup(text, input_filename)
        return self.parser.parse(None, lexer=self.lexer.lexer)

    # operator precedence (lower in the list = higher priority)
    precedence = (
        ("left", "COMMA"),
        ("right", "TERNARY_OPEN", "COLON"),
        ("left", "LOGICAL_OR"),
        ("left", "LOGICAL_AND"),
        ("left", "OR"),
        ("left", "XOR"),
        ("left", "AND"),
        ("left", "COMP_EQ", "COMP_NEQ", "COMP_LE", "COMP_GE", "COMP_LT", "COMP_GT"),
        ("left", "SHIFT_LEFT", "SHIFT_RIGHT", "SHIFTU_RIGHT"),
        ("left", "PLUS", "MINUS"),
        ("left", "TIMES", "DIVIDE", "MODULO"),
        ("left", "LOGICAL_NOT", "BINARY_NOT"),
    )

    def p_error(self, t):
        if t is None:
            raise generic.ScriptError("Syntax error, unexpected end-of-file")
        else:
            raise generic.ScriptError('Syntax error, unexpected token "{}"'.format(t.value), t.lineno)

    #
    # Main script blocks
    #
    def p_main_script(self, t):
        "main_script : script"
        t[0] = general.MainScript(t[1])

    def p_script(self, t):
        """script :
        | script main_block"""
        if len(t) == 1:
            t[0] = []
        else:
            t[0] = t[1] + [t[2]]

    def p_main_block(self, t):
        """main_block : switch
        | random_switch
        | produce
        | spriteset
        | spritegroup
        | spritelayout
        | template_declaration
        | tilelayout
        | town_names
        | cargotable
        | railtype
        | roadtype
        | tramtype
        | grf_block
        | param_assignment
        | skip_all
        | conditional
        | loop
        | item
        | property_block
        | graphics_block
        | liveryoverride_block
        | error_block
        | disable_item
        | deactivate
        | replace
        | replace_new
        | base_graphics
        | font_glyph
        | alt_sprites
        | snowline
        | engine_override
        | sort_vehicles
        | basecost
        | constant"""
        t[0] = t[1]

    #
    # Expressions
    #
    def p_expression(self, t):
        """expression : NUMBER
        | FLOAT
        | param
        | variable
        | ID
        | STRING_LITERAL
        | string"""
        t[0] = t[1]

    def p_parenthesed_expression(self, t):
        "expression : LPAREN expression RPAREN"
        t[0] = t[2]

    def p_parameter(self, t):
        "param : PARAMETER LBRACKET expression RBRACKET"
        t[0] = expression.Parameter(t[3], t.lineno(1), True)

    def p_parameter_other_grf(self, t):
        "param : PARAMETER LBRACKET expression COMMA expression RBRACKET"
        t[0] = expression.OtherGRFParameter(t[3], t[5], t.lineno(1))

    code_to_op = {
        "+": nmlop.ADD,
        "-": nmlop.SUB,
        "*": nmlop.MUL,
        "/": nmlop.DIV,
        "%": nmlop.MOD,
        "&": nmlop.AND,
        "|": nmlop.OR,
        "^": nmlop.XOR,
        "&&": nmlop.AND,
        "||": nmlop.OR,
        "==": nmlop.CMP_EQ,
        "!=": nmlop.CMP_NEQ,
        "<=": nmlop.CMP_LE,
        ">=": nmlop.CMP_GE,
        "<": nmlop.CMP_LT,
        ">": nmlop.CMP_GT,
        "<<": nmlop.SHIFT_LEFT,
        ">>": nmlop.SHIFT_RIGHT,
        ">>>": nmlop.SHIFTU_RIGHT,
    }

    def p_binop(self, t):
        """expression : expression PLUS expression
        | expression MINUS expression
        | expression TIMES expression
        | expression DIVIDE expression
        | expression MODULO expression
        | expression AND expression
        | expression OR expression
        | expression XOR expression
        | expression SHIFT_LEFT expression
        | expression SHIFT_RIGHT expression
        | expression SHIFTU_RIGHT expression
        | expression COMP_EQ expression
        | expression COMP_NEQ expression
        | expression COMP_LE expression
        | expression COMP_GE expression
        | expression COMP_LT expression
        | expression COMP_GT expression"""
        t[0] = expression.BinOp(self.code_to_op[t[2]], t[1], t[3], t[1].pos)

    def p_binop_logical(self, t):
        """expression : expression LOGICAL_AND expression
        | expression LOGICAL_OR expression"""
        t[0] = expression.BinOp(self.code_to_op[t[2]], expression.Boolean(t[1]), expression.Boolean(t[3]), t[1].pos)

    def p_logical_not(self, t):
        "expression : LOGICAL_NOT expression"
        t[0] = expression.Not(expression.Boolean(t[2]), t.lineno(1))

    def p_binary_not(self, t):
        "expression : BINARY_NOT expression"
        t[0] = expression.BinNot(t[2], t.lineno(1))

    def p_ternary_op(self, t):
        "expression : expression TERNARY_OPEN expression COLON expression"
        t[0] = expression.TernaryOp(t[1], t[3], t[5], t[1].pos)

    def p_unary_minus(self, t):
        "expression : MINUS expression"
        t[0] = nmlop.SUB(0, t[2], t.lineno(1))

    def p_variable(self, t):
        "variable : VARIABLE LBRACKET expression_list RBRACKET"
        t[0] = expression.Variable(*t[3])
        t[0].pos = t.lineno(1)

    def p_function(self, t):
        "expression : ID LPAREN expression_list RPAREN"
        t[0] = expression.FunctionCall(t[1], t[3], t[1].pos)

    def p_array(self, t):
        "expression : LBRACKET expression_list RBRACKET"
        t[0] = expression.Array(t[2], t.lineno(1))

    #
    # Commonly used non-terminals that are not expressions
    #
    def p_assignment_list(self, t):
        """assignment_list : assignment
        | param_desc
        | assignment_list assignment
        | assignment_list param_desc"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    def p_assignment(self, t):
        "assignment : ID COLON expression SEMICOLON"
        t[0] = assignment.Assignment(t[1], t[3], t[1].pos)

    def p_param_desc(self, t):
        """param_desc : PARAMETER expression LBRACE setting_list RBRACE
        | PARAMETER LBRACE setting_list RBRACE"""
        if len(t) == 5:
            t[0] = grf.ParameterDescription(t[3])
        else:
            t[0] = grf.ParameterDescription(t[4], t[2])

    def p_setting_list(self, t):
        """setting_list : setting
        | setting_list setting"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    def p_setting(self, t):
        "setting : ID LBRACE setting_value_list RBRACE"
        t[0] = grf.ParameterSetting(t[1], t[3])

    def p_setting_value_list(self, t):
        """setting_value_list : setting_value
        | setting_value_list setting_value"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    def p_setting_value(self, t):
        "setting_value : assignment"
        t[0] = t[1]

    def p_names_setting_value(self, t):
        "setting_value : ID COLON LBRACE name_string_list RBRACE SEMICOLON"
        t[0] = assignment.Assignment(t[1], t[4], t[1].pos)

    def p_name_string_list(self, t):
        """name_string_list : name_string_item
        | name_string_list name_string_item"""
        if len(t) == 2:
            t[0] = expression.Array([t[1]], t[1].pos)
        else:
            t[0] = expression.Array(t[1].values + [t[2]], t[1].pos)

    def p_name_string_item(self, t):
        "name_string_item : expression COLON string SEMICOLON"
        t[0] = assignment.Assignment(t[1], t[3], t[1].pos)

    def p_string(self, t):
        "string : STRING LPAREN expression_list RPAREN"
        t[0] = expression.String(t[3], t.lineno(1))

    def p_non_empty_expression_list(self, t):
        """non_empty_expression_list : expression
        | non_empty_expression_list COMMA expression"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[3]]

    def p_expression_list(self, t):
        """expression_list :
        | non_empty_expression_list
        | non_empty_expression_list COMMA"""
        t[0] = [] if len(t) == 1 else t[1]

    def p_non_empty_id_list(self, t):
        """non_empty_id_list : ID
        | non_empty_id_list COMMA ID"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[3]]

    def p_id_list(self, t):
        """id_list :
        | non_empty_id_list
        | non_empty_id_list COMMA"""
        t[0] = [] if len(t) == 1 else t[1]

    def p_generic_assignment(self, t):
        "generic_assignment : expression COLON expression SEMICOLON"
        t[0] = assignment.Assignment(t[1], t[3], t.lineno(1))

    def p_generic_assignment_list(self, t):
        """generic_assignment_list :
        | generic_assignment_list generic_assignment"""
        t[0] = [] if len(t) == 1 else t[1] + [t[2]]

    def p_snowline_assignment(self, t):
        """snowline_assignment : expression COLON expression SEMICOLON
        | expression COLON expression UNIT SEMICOLON"""
        unit_value = None if len(t) == 5 else unit.get_unit(t[4])
        t[0] = assignment.UnitAssignment(t[1], t[3], unit_value, t.lineno(1))

    def p_snowline_assignment_list(self, t):
        """snowline_assignment_list :
        | snowline_assignment_list snowline_assignment"""
        t[0] = [] if len(t) == 1 else t[1] + [t[2]]

    #
    # Item blocks
    #
    def p_item(self, t):
        "item : ITEM LPAREN expression_list RPAREN LBRACE script RBRACE"
        t[0] = item.Item(t[3], t[6], t.lineno(1))

    def p_property_block(self, t):
        "property_block : PROPERTY LBRACE property_list RBRACE"
        t[0] = item.PropertyBlock(t[3], t.lineno(1))

    def p_property_list(self, t):
        """property_list : property_assignment
        | property_list property_assignment"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    def p_property_assignment(self, t):
        """property_assignment : ID COLON expression SEMICOLON
        | ID COLON expression UNIT SEMICOLON
        | NUMBER COLON expression SEMICOLON
        | NUMBER COLON expression UNIT SEMICOLON"""
        name = t[1]
        unit_value = None if len(t) == 5 else unit.get_unit(t[4])
        t[0] = item.Property(name, t[3], unit_value, t.lineno(1))

    def p_graphics_block(self, t):
        "graphics_block : GRAPHICS LBRACE graphics_list RBRACE"
        t[0] = item.GraphicsBlock(t[3][0], t[3][1], t.lineno(1))

    def p_liveryoverride_block(self, t):
        "liveryoverride_block : LIVERYOVERRIDE LPAREN expression RPAREN LBRACE graphics_list RBRACE"
        t[0] = item.LiveryOverride(t[3], item.GraphicsBlock(t[6][0], t[6][1], t.lineno(1)), t.lineno(1))

    def p_graphics_list(self, t):
        """graphics_list : graphics_assignment_list
        | graphics_assignment_list switch_value
        | switch_value"""
        # Save graphics block as a tuple, we need to add position info later
        if len(t) == 2:
            if isinstance(t[1], list):
                t[0] = (t[1], None)
            else:
                t[0] = ([], t[1])
        else:
            t[0] = (t[1], t[2])

    def p_graphics_assignment(self, t):
        "graphics_assignment : expression COLON switch_value"
        t[0] = item.GraphicsDefinition(t[1], t[3])

    def p_graphics_assignment_list(self, t):
        """graphics_assignment_list : graphics_assignment
        | graphics_assignment_list graphics_assignment"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    #
    # Program flow control (if/else/while)
    #
    def p_conditional(self, t):
        """conditional : if_else_parts
        | if_else_parts else_block"""
        parts = t[1]
        if len(t) > 2:
            parts.append(t[2])
        t[0] = conditional.ConditionalList(parts)

    def p_else_block(self, t):
        "else_block : ELSE LBRACE script RBRACE"
        t[0] = conditional.Conditional(None, t[3], t.lineno(1))

    def p_if_else_parts(self, t):
        """if_else_parts : IF LPAREN expression RPAREN LBRACE script RBRACE
        | if_else_parts ELSE IF LPAREN expression RPAREN LBRACE script RBRACE"""
        if len(t) == 8:
            t[0] = [conditional.Conditional(t[3], t[6], t.lineno(1))]
        else:
            t[0] = t[1] + [conditional.Conditional(t[5], t[8], t.lineno(2))]

    def p_loop(self, t):
        "loop : WHILE LPAREN expression RPAREN LBRACE script RBRACE"
        t[0] = loop.Loop(t[3], t[6], t.lineno(1))

    #
    # (Random) Switch block
    #
    def p_switch(self, t):
        "switch : SWITCH LPAREN expression_list RPAREN LBRACE switch_body RBRACE"
        t[0] = switch.Switch(t[3], t[6], t.lineno(1))

    def p_switch_body(self, t):
        """switch_body : switch_ranges switch_value
        | switch_ranges"""
        t[0] = switch.SwitchBody(t[1], t[2] if len(t) == 3 else None)

    def p_switch_ranges(self, t):
        """switch_ranges :
        | switch_ranges expression COLON switch_value
        | switch_ranges expression UNIT COLON switch_value
        | switch_ranges expression RANGE expression COLON switch_value
        | switch_ranges expression RANGE expression UNIT COLON switch_value"""
        if len(t) == 1:
            t[0] = []
        elif len(t) == 5:
            t[0] = t[1] + [switch.SwitchRange(t[2], t[2], t[4])]
        elif len(t) == 6:
            t[0] = t[1] + [switch.SwitchRange(t[2], t[2], t[5], t[3])]
        elif len(t) == 7:
            t[0] = t[1] + [switch.SwitchRange(t[2], t[4], t[6])]
        else:
            t[0] = t[1] + [switch.SwitchRange(t[2], t[4], t[7], t[5])]

    def p_switch_value(self, t):
        """switch_value : RETURN expression SEMICOLON
        | RETURN SEMICOLON
        | expression SEMICOLON"""
        if len(t) == 4:
            t[0] = switch.SwitchValue(t[2], True, t[2].pos)
        elif t[1] == "return":
            t[0] = switch.SwitchValue(None, True, t.lineno(1))
        else:
            t[0] = switch.SwitchValue(t[1], False, t[1].pos)

    def p_random_switch(self, t):
        "random_switch : RANDOMSWITCH LPAREN expression_list RPAREN LBRACE random_body RBRACE"
        t[0] = switch.RandomSwitch(t[3], t[6], t.lineno(1))

    def p_random_body(self, t):
        """random_body :
        | random_body expression COLON switch_value"""
        if len(t) == 1:
            t[0] = []
        else:
            t[0] = t[1] + [switch.RandomChoice(t[2], t[4])]

    def p_produce_cargo_list(self, t):
        """produce_cargo_list : LBRACKET RBRACKET
        | LBRACKET setting_value_list RBRACKET"""
        if len(t) == 3:
            t[0] = []
        else:
            t[0] = t[2]

    def p_produce(self, t):
        """produce : PRODUCE LPAREN ID COMMA expression_list RPAREN SEMICOLON
        | PRODUCE LPAREN ID COMMA produce_cargo_list COMMA produce_cargo_list COMMA expression RPAREN
        | PRODUCE LPAREN ID COMMA produce_cargo_list COMMA produce_cargo_list RPAREN"""
        if len(t) == 8:
            t[0] = produce.ProduceOld([t[3]] + t[5], t.lineno(1))
        elif len(t) == 11:
            t[0] = produce.Produce(t[3], t[5], t[7], t[9], t.lineno(1))
        else:
            t[0] = produce.Produce(t[3], t[5], t[7], expression.ConstantNumeric(0), t.lineno(1))

    #
    # Real sprites and related stuff
    #
    def p_real_sprite(self, t):
        """real_sprite : LBRACKET expression_list RBRACKET
        | ID COLON LBRACKET expression_list RBRACKET"""
        if len(t) == 4:
            t[0] = real_sprite.RealSprite(param_list=t[2], poslist=[t.lineno(1)])
        else:
            t[0] = real_sprite.RealSprite(param_list=t[4], label=t[1], poslist=[t.lineno(1)])

    def p_recolour_assignment_list(self, t):
        """recolour_assignment_list :
        | recolour_assignment_list recolour_assignment"""
        t[0] = [] if len(t) == 1 else t[1] + [t[2]]

    def p_recolour_assignment_1(self, t):
        "recolour_assignment : expression COLON expression SEMICOLON"
        t[0] = assignment.Assignment(assignment.Range(t[1], None), assignment.Range(t[3], None), t[1].pos)

    def p_recolour_assignment_2(self, t):
        "recolour_assignment : expression RANGE expression COLON expression RANGE expression SEMICOLON"
        t[0] = assignment.Assignment(assignment.Range(t[1], t[3]), assignment.Range(t[5], t[7]), t[1].pos)

    def p_recolour_assignment_3(self, t):
        "recolour_assignment : expression RANGE expression COLON expression SEMICOLON"
        t[0] = assignment.Assignment(assignment.Range(t[1], t[3]), assignment.Range(t[5], None), t[1].pos)

    def p_recolour_sprite(self, t):
        """real_sprite : RECOLOUR_SPRITE LBRACE recolour_assignment_list RBRACE
        | ID COLON RECOLOUR_SPRITE LBRACE recolour_assignment_list RBRACE"""
        if len(t) == 5:
            t[0] = real_sprite.RecolourSprite(mapping=t[3], poslist=[t.lineno(1)])
        else:
            t[0] = real_sprite.RecolourSprite(mapping=t[5], label=t[1], poslist=[t.lineno(1)])

    def p_template_declaration(self, t):
        "template_declaration : TEMPLATE ID LPAREN id_list RPAREN LBRACE spriteset_contents RBRACE"
        t[0] = spriteblock.TemplateDeclaration(t[2], t[4], t[7], t.lineno(1))

    def p_template_usage(self, t):
        """template_usage : ID LPAREN expression_list RPAREN
        | ID COLON ID LPAREN expression_list RPAREN"""
        if len(t) == 5:
            t[0] = real_sprite.TemplateUsage(t[1], t[3], None, t.lineno(1))
        else:
            t[0] = real_sprite.TemplateUsage(t[3], t[5], t[1], t.lineno(1))

    def p_spriteset_contents(self, t):
        """spriteset_contents : real_sprite
        | template_usage
        | spriteset_contents real_sprite
        | spriteset_contents template_usage"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    def p_replace(self, t):
        """replace : REPLACESPRITE LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE
        | REPLACESPRITE ID LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE"""
        if len(t) == 9:
            t[0] = replace.ReplaceSprite(t[4], t[7], t[2], t.lineno(1))
        else:
            t[0] = replace.ReplaceSprite(t[3], t[6], None, t.lineno(1))

    def p_replace_new(self, t):
        """replace_new : REPLACENEWSPRITE LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE
        | REPLACENEWSPRITE ID LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE"""
        if len(t) == 9:
            t[0] = replace.ReplaceNewSprite(t[4], t[7], t[2], t.lineno(1))
        else:
            t[0] = replace.ReplaceNewSprite(t[3], t[6], None, t.lineno(1))

    def p_base_graphics(self, t):
        """base_graphics : BASE_GRAPHICS LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE
        | BASE_GRAPHICS ID LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE"""
        if len(t) == 9:
            t[0] = base_graphics.BaseGraphics(t[4], t[7], t[2], t.lineno(1))
        else:
            t[0] = base_graphics.BaseGraphics(t[3], t[6], None, t.lineno(1))

    def p_font_glyph(self, t):
        """font_glyph : FONTGLYPH LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE
        | FONTGLYPH ID LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE"""
        if len(t) == 9:
            t[0] = font.FontGlyphBlock(t[4], t[7], t[2], t.lineno(1))
        else:
            t[0] = font.FontGlyphBlock(t[3], t[6], None, t.lineno(1))

    def p_alt_sprites(self, t):
        "alt_sprites : ALT_SPRITES LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE"
        t[0] = alt_sprites.AltSpritesBlock(t[3], t[6], t.lineno(1))

    #
    # Sprite sets/groups and such
    #

    def p_spriteset(self, t):
        "spriteset : SPRITESET LPAREN expression_list RPAREN LBRACE spriteset_contents RBRACE"
        t[0] = spriteblock.SpriteSet(t[3], t[6], t.lineno(1))

    def p_spritegroup_normal(self, t):
        "spritegroup : SPRITEGROUP ID LBRACE spriteview_list RBRACE"
        t[0] = spriteblock.SpriteGroup(t[2], t[4], t.lineno(1))

    def p_spritelayout(self, t):
        """spritelayout : SPRITELAYOUT ID LBRACE layout_sprite_list RBRACE
        | SPRITELAYOUT ID LPAREN id_list RPAREN LBRACE layout_sprite_list RBRACE"""
        if len(t) == 6:
            t[0] = spriteblock.SpriteLayout(t[2], [], t[4], t.lineno(1))
        else:
            t[0] = spriteblock.SpriteLayout(t[2], t[4], t[7], t.lineno(1))

    def p_spriteview_list(self, t):
        """spriteview_list :
        | spriteview_list spriteview"""
        if len(t) == 1:
            t[0] = []
        else:
            t[0] = t[1] + [t[2]]

    def p_spriteview(self, t):
        """spriteview : ID COLON LBRACKET expression_list RBRACKET SEMICOLON
        | ID COLON expression SEMICOLON"""
        if len(t) == 7:
            t[0] = spriteblock.SpriteView(t[1], t[4], t.lineno(1))
        else:
            t[0] = spriteblock.SpriteView(t[1], [t[3]], t.lineno(1))

    def p_layout_sprite_list(self, t):
        """layout_sprite_list :
        | layout_sprite_list layout_sprite"""
        if len(t) == 1:
            t[0] = []
        else:
            t[0] = t[1] + [t[2]]

    def p_layout_sprite(self, t):
        "layout_sprite : ID LBRACE layout_param_list RBRACE"
        t[0] = spriteblock.LayoutSprite(t[1], t[3], t.lineno(1))

    def p_layout_param_list(self, t):
        """layout_param_list : assignment
        | layout_param_list assignment"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    #
    # Town names
    #
    def p_town_names(self, t):
        """town_names : TOWN_NAMES LPAREN expression RPAREN LBRACE town_names_param_list RBRACE
        | TOWN_NAMES LBRACE town_names_param_list RBRACE"""
        if len(t) == 8:
            t[0] = townnames.TownNames(t[3], t[6], t.lineno(1))
        else:
            t[0] = townnames.TownNames(None, t[3], t.lineno(1))

    def p_town_names_param_list(self, t):
        """town_names_param_list : town_names_param
        | town_names_param_list town_names_param"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    def p_town_names_param(self, t):
        """town_names_param : ID COLON string SEMICOLON
        | LBRACE town_names_part_list RBRACE
        | LBRACE town_names_part_list COMMA RBRACE"""
        if t[1] != "{":
            t[0] = townnames.TownNamesParam(t[1], t[3], t.lineno(1))
        else:
            t[0] = townnames.TownNamesPart(t[2], t.lineno(1))

    def p_town_names_part_list(self, t):
        """town_names_part_list : town_names_part
        | town_names_part_list COMMA town_names_part"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[3]]

    def p_town_names_part(self, t):
        """town_names_part : TOWN_NAMES LPAREN expression COMMA expression RPAREN
        | ID LPAREN STRING_LITERAL COMMA expression RPAREN"""
        if t[1] == "town_names":
            t[0] = townnames.TownNamesEntryDefinition(t[3], t[5], t.lineno(1))
        else:
            t[0] = townnames.TownNamesEntryText(t[1], t[3], t[5], t.lineno(1))

    #
    # Snow line
    #
    def p_snowline(self, t):
        "snowline : SNOWLINE LPAREN ID RPAREN LBRACE snowline_assignment_list RBRACE"
        t[0] = snowline.Snowline(t[3], t[6], t.lineno(1))

    #
    # Various misc. main script blocks that don't belong anywhere else
    #
    def p_param_assignment(self, t):
        "param_assignment : expression EQ expression SEMICOLON"
        t[0] = actionD.ParameterAssignment(t[1], t[3])

    def p_error_block(self, t):
        "error_block : ERROR LPAREN expression_list RPAREN SEMICOLON"
        t[0] = error.Error(t[3], t.lineno(1))

    def p_disable_item(self, t):
        "disable_item : DISABLE_ITEM LPAREN expression_list RPAREN SEMICOLON"
        t[0] = disable_item.DisableItem(t[3], t.lineno(1))

    def p_cargotable(self, t):
        """cargotable : CARGOTABLE LBRACE cargotable_list RBRACE
        | CARGOTABLE LBRACE cargotable_list COMMA RBRACE"""
        t[0] = cargotable.CargoTable(t[3], t.lineno(1))

    def p_cargotable_list(self, t):
        """cargotable_list : ID
        | STRING_LITERAL
        | cargotable_list COMMA ID
        | cargotable_list COMMA STRING_LITERAL"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[3]]

    def p_railtypetable(self, t):
        """railtype : RAILTYPETABLE LBRACE tracktypetable_list RBRACE
        | RAILTYPETABLE LBRACE tracktypetable_list COMMA RBRACE"""
        t[0] = tracktypetable.RailtypeTable(t[3], t.lineno(1))

    def p_roadtypetable(self, t):
        """roadtype : ROADTYPETABLE LBRACE tracktypetable_list RBRACE
        | ROADTYPETABLE LBRACE tracktypetable_list COMMA RBRACE"""
        t[0] = tracktypetable.RoadtypeTable(t[3], t.lineno(1))

    def p_tramtypetable(self, t):
        """tramtype : TRAMTYPETABLE LBRACE tracktypetable_list RBRACE
        | TRAMTYPETABLE LBRACE tracktypetable_list COMMA RBRACE"""
        t[0] = tracktypetable.TramtypeTable(t[3], t.lineno(1))

    def p_tracktypetable_list(self, t):
        """tracktypetable_list : tracktypetable_item
        | tracktypetable_list COMMA tracktypetable_item"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[3]]

    def p_tracktypetable_item(self, t):
        """tracktypetable_item : ID
        | STRING_LITERAL
        | ID COLON LBRACKET expression_list RBRACKET"""
        if len(t) == 2:
            t[0] = t[1]
        else:
            t[0] = assignment.Assignment(t[1], t[4], t[1].pos)

    def p_basecost(self, t):
        "basecost : BASECOST LBRACE generic_assignment_list RBRACE"
        t[0] = basecost.BaseCost(t[3], t.lineno(1))

    def p_deactivate(self, t):
        "deactivate : DEACTIVATE LPAREN expression_list RPAREN SEMICOLON"
        t[0] = deactivate.DeactivateBlock(t[3], t.lineno(1))

    def p_grf_block(self, t):
        "grf_block : GRF LBRACE assignment_list RBRACE"
        t[0] = grf.GRF(t[3], t.lineno(1))

    def p_skip_all(self, t):
        "skip_all : SKIP_ALL SEMICOLON"
        t[0] = skipall.SkipAll(t.lineno(1))

    def p_engine_override(self, t):
        "engine_override : ENGINE_OVERRIDE LPAREN expression_list RPAREN SEMICOLON"
        t[0] = override.EngineOverride(t[3], t.lineno(1))

    def p_sort_vehicles(self, t):
        "sort_vehicles : SORT_VEHICLES LPAREN expression_list RPAREN SEMICOLON"
        t[0] = sort_vehicles.SortVehicles(t[3], t.lineno(1))

    def p_tilelayout(self, t):
        "tilelayout : TILELAYOUT ID LBRACE tilelayout_list RBRACE"
        t[0] = tilelayout.TileLayout(t[2], t[4], t.lineno(1))

    def p_tilelayout_list(self, t):
        """tilelayout_list : tilelayout_item
        | tilelayout_list tilelayout_item"""
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1] + [t[2]]

    def p_tilelayout_item_tile(self, t):
        "tilelayout_item : expression COMMA expression COLON expression SEMICOLON"
        t[0] = tilelayout.LayoutTile(t[1], t[3], t[5])

    def p_tilelayout_item_prop(self, t):
        "tilelayout_item : assignment"
        t[0] = t[1]

    def p_constant(self, t):
        "constant : CONST expression EQ expression SEMICOLON"
        t[0] = constant.Constant(t[2], t[4])



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""


class BaseAction:
    def prepare_output(self, sprite_num):
        """
        Called just before L{write}, this function can be used to do some
        last modifications to the action (like resolving some IDs that can't
        be resolved earlier).

        @param sprite_num: The sprite number in the nfo/grf output.
        @type  sprite_num: C{int}
        """
        pass

    def write(self, file):
        """
        Write this action to the given outputfile.

        @param file: The outputfile to write the data to.
        @type  file: L{SpriteOutputBase}
        """
        raise NotImplementedError("write is not implemented in {!r}".format(type(self)))

    def skip_action7(self):
        """
        Can this action be skipped safely by an action7?

        @return True iff this action can be skipped by action7.
        """
        return True

    def skip_action9(self):
        """
        Can this action be skipped safely by an action9?

        @return True iff this action can be skipped by action9.
        """
        return True

    def skip_needed(self):
        """
        Do we need to skip this action at all?

        @return False when it doesn't matter whether this action is skipped
            or not.
        """
        return True



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""


class SpriteCountAction:
    def __init__(self, count):
        self.count = count

    def prepare_output(self, sprite_num):
        assert sprite_num == 0

    def write(self, file):
        file.start_sprite(4)
        file.print_dword(self.count)
        file.newline()
        file.end_sprite()



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import free_number_list, generic
from nml.actions import base_action

free_parameters = free_number_list.FreeNumberList(
    list(range(0x40, 0x80)),
    "No free parameters available to use for internal computations.",
    "No unique free parameters available for internal computations.",
)


def print_stats():
    """
    Print statistics about used ids.
    """
    if free_parameters.stats[0] > 0:
        generic.print_info(
            "Concurrent ActionD registers: {}/{} ({})".format(
                free_parameters.stats[0], free_parameters.total_amount, str(free_parameters.stats[1])
            )
        )


class Action6(base_action.BaseAction):
    def __init__(self):
        self.modifications = []

    def modify_bytes(self, param, num_bytes, offset):
        self.modifications.append((param, num_bytes, offset))

    def write(self, file):
        size = 2 + 5 * len(self.modifications)
        file.start_sprite(size)
        file.print_bytex(6)
        file.newline()
        for mod in self.modifications:
            file.print_bytex(mod[0])
            file.print_bytex(mod[1])
            file.print_bytex(0xFF)
            file.print_wordx(mod[2])
            file.newline()
        file.print_bytex(0xFF)
        file.newline()
        file.end_sprite()

    def skip_action7(self):
        return False



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic

from .base_expression import ConstantNumeric, Expression, Type


class Variable(Expression):
    def __init__(self, num, shift=None, mask=None, param=None, pos=None):
        Expression.__init__(self, pos)
        self.num = num
        self.shift = shift if shift is not None else ConstantNumeric(0)
        self.mask = mask if mask is not None else ConstantNumeric(0xFFFFFFFF)
        self.param = param
        self.add = None
        self.div = None
        self.mod = None
        self.extra_params = []

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Action2 variable")
        self.num.debug_print(indentation + 2)
        if self.param is not None:
            generic.print_dbg(indentation + 2, "Parameter:")
            if isinstance(self.param, str):
                generic.print_dbg(indentation + 4, "Procedure call:", self.param)
            else:
                self.param.debug_print(indentation + 4)
            if len(self.extra_params) > 0:
                generic.print_dbg(indentation + 2, "Extra parameters:")
            for extra_param in self.extra_params:
                extra_param.debug_print(indentation + 4)

    def __str__(self):
        num = "0x{:02X}".format(self.num.value) if isinstance(self.num, ConstantNumeric) else str(self.num)
        ret = "var[{}, {}, {}".format(num, self.shift, self.mask)
        if self.param is not None:
            ret += ", {}".format(self.param)
        ret += "]"
        if self.add is not None:
            ret = "({} + {})".format(ret, self.add)
        if self.div is not None:
            ret = "({} / {})".format(ret, self.div)
        if self.mod is not None:
            ret = "({} % {})".format(ret, self.mod)
        return ret

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        num = self.num.reduce(id_dicts)
        shift = self.shift.reduce(id_dicts)
        mask = self.mask.reduce(id_dicts)
        param = self.param.reduce(id_dicts) if self.param is not None else None
        if (
            num.type() != Type.INTEGER
            or shift.type() != Type.INTEGER
            or mask.type() != Type.INTEGER
            or (param is not None and param.type() != Type.INTEGER)
        ):
            raise generic.ScriptError("All parts of a variable access must be integers.", self.pos)
        var = Variable(num, shift, mask, param, self.pos)
        var.add = None if self.add is None else self.add.reduce(id_dicts)
        var.div = None if self.div is None else self.div.reduce(id_dicts)
        var.mod = None if self.mod is None else self.mod.reduce(id_dicts)
        var.extra_params = [(extra_param[0], extra_param[1].reduce(id_dicts)) for extra_param in self.extra_params]
        return var

    def supported_by_action2(self, raise_error):
        return True

    def supported_by_actionD(self, raise_error):
        if raise_error:
            if isinstance(self.num, ConstantNumeric):
                if self.num.value == 0x7C:
                    raise generic.ScriptError("LOAD_PERM is only available in switch-blocks.", self.pos)
                if self.num.value == 0x7D:
                    raise generic.ScriptError("LOAD_TEMP is only available in switch-blocks.", self.pos)
            raise generic.ScriptError("Variable accesses are not supported outside of switch-blocks.", self.pos)
        return False



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants, nmlop
from nml.actions import action6, action7, base_action
from nml.ast import base_statement, conditional


class ActionD(base_action.BaseAction):
    """
    ActionD class
    General procedure: target = param1 op param2
    If one of the params is 0xFF, the value of 'data' is read instead.

    @ivar target: Number of the target parameter
    @ivar target: L{ConstantNumeric}

    @ivar param1: Parameter number of the first operand
    @type param1: L{ConstantNumeric}

    @ivar op: (Binary) operator to use.
    @type op: L{Operator}

    @ivar param2: Parameter number of the second operand
    @type param2: L{ConstantNumeric}

    @ivar data: Numerical data that will be used instead of parameter value,
                    if the parameter number is 0xFF. None if n/a.
    @type data: L{ConstantNumeric} or C{None}
    """

    def __init__(self, target, param1, op, param2, data=None):
        self.target = target
        self.param1 = param1
        self.op = op
        self.param2 = param2
        self.data = data

    def write(self, file):
        size = 5
        if self.data is not None:
            size += 4

        # print the statement for easier debugging
        str1 = "param[{}]".format(self.param1) if self.param1.value != 0xFF else str(self.data)
        str2 = "param[{}]".format(self.param2) if self.param2.value != 0xFF else str(self.data)
        str_total = self.op.to_string(str1, str2) if self.op != nmlop.ASSIGN else str1
        file.comment("param[{}] = {}".format(self.target, str_total))

        file.start_sprite(size)
        file.print_bytex(0x0D)
        self.target.write(file, 1)
        file.print_bytex(self.op.actd_num, self.op.actd_str)
        self.param1.write(file, 1)
        self.param2.write(file, 1)
        if self.data is not None:
            self.data.write(file, 4)
        file.newline()
        file.end_sprite()

    def skip_action7(self):
        return False


class ParameterAssignment(base_statement.BaseStatement):
    """
    AST-node for a parameter assignment.
    NML equivalent: param[$num] = $expr;

    @ivar param: Target expression to assign (must evaluate to a parameter)
    @type param: L{Expression}

    @ivar value: Value to assign to this parameter
    @type value: L{Expression}
    """

    def __init__(self, param, value):
        base_statement.BaseStatement.__init__(self, "parameter assignment", param.pos)
        self.param = param
        self.value = value

    def register_names(self):
        self.param = self.param.reduce(global_constants.const_list, unknown_id_fatal=False)
        if isinstance(self.param, expression.Identifier):
            if global_constants.identifier_refcount[self.param.value] == 0:
                generic.print_warning(
                    generic.Warning.OPTIMISATION,
                    "Named parameter '{}' is not referenced, ignoring.".format(self.param.value),
                    self.param.pos,
                )
                return
            num = action6.free_parameters.pop_unique(self.pos)
            global_constants.named_parameters[self.param.value] = num

    def pre_process(self):
        self.value = self.value.reduce(global_constants.const_list)

        if isinstance(self.param, expression.SpecialParameter):
            if not self.param.can_assign():
                raise generic.ScriptError(
                    "Trying to assign a value to the read-only variable '{}'".format(self.param.name), self.param.pos
                )
        elif isinstance(self.param, expression.Identifier):
            return
        elif not isinstance(self.param, expression.Parameter):
            raise generic.ScriptError("Left side of an assignment must be a parameter.", self.param.pos)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Parameter assignment")
        self.param.debug_print(indentation + 2)
        self.value.debug_print(indentation + 2)

    def get_action_list(self):
        return parse_actionD(self)

    def __str__(self):
        return "{} = {};\n".format(self.param, self.value)


# prevent evaluating common sub-expressions multiple times
def parse_subexpression(expr, action_list):
    if isinstance(expr, expression.ConstantNumeric) or (
        isinstance(expr, expression.Parameter) and isinstance(expr.num, expression.ConstantNumeric)
    ):
        return expr
    else:
        tmp_param, tmp_param_actions = get_tmp_parameter(expr)
        action_list.extend(tmp_param_actions)
        return expression.Parameter(expression.ConstantNumeric(tmp_param))


# returns a (param_num, action_list) tuple.
def get_tmp_parameter(expr):
    param = action6.free_parameters.pop(expr.pos)
    actions = parse_actionD(ParameterAssignment(expression.Parameter(expression.ConstantNumeric(param)), expr))
    return (param, actions)


def write_action_value(expr, action_list, act6, offset, size):
    """
    Write a single numeric value that is part of an action
    Possibly use action6 and/or actionD if needed

    @param expr: Expression to write
    @type expr: L{Expression}

    @param action_list: Action list to append any extra actions to
    @type action_list: C{list} of L{BaseAction}

    @param act6: Action6 to append any modifications to
    @type act6: L{Action6}

    @param offset: Offset to use for the action 6 (if applicable)
    @type offset: C{int}

    @param size: Size to use for the action 6 (if applicable)
    @type size: C{int}

    @return: 2-tuple containing -
        - the value to be used when writing the action (0 if overwritten by action 6)
        - the offset just past this numeric value (=offset + size)
    @rtype: C{tuple} of (L{ConstantNumeric}, C{int})
    """
    if isinstance(expr, expression.ConstantNumeric):
        result = expr
    elif isinstance(expr, expression.Parameter) and isinstance(expr.num, expression.ConstantNumeric):
        act6.modify_bytes(expr.num.value, size, offset)
        result = expression.ConstantNumeric(0)
    else:
        tmp_param, tmp_param_actions = get_tmp_parameter(expr)
        action_list.extend(tmp_param_actions)
        act6.modify_bytes(tmp_param, size, offset)
        result = expression.ConstantNumeric(0)
    return result, offset + size


def parse_ternary_op(assignment):
    assert isinstance(assignment.value, expression.TernaryOp)
    actions = parse_actionD(ParameterAssignment(assignment.param, assignment.value.expr2))
    cond_block = conditional.Conditional(
        assignment.value.guard, [ParameterAssignment(assignment.param, assignment.value.expr1)], None
    )
    actions.extend(conditional.ConditionalList([cond_block]).get_action_list())
    return actions


def parse_abs_op(assignment):
    assert isinstance(assignment.value, expression.AbsOp)
    actions = parse_actionD(ParameterAssignment(assignment.param, assignment.value.expr))
    cond_block = conditional.Conditional(
        nmlop.CMP_LT(assignment.value.expr, 0),
        [ParameterAssignment(assignment.param, nmlop.SUB(0, assignment.value.expr))],
        None,
    )
    actions.extend(conditional.ConditionalList([cond_block]).get_action_list())
    return actions


def parse_special_check(assignment):
    check = assignment.value
    assert isinstance(check, expression.SpecialCheck)
    actions = parse_actionD(ParameterAssignment(assignment.param, expression.ConstantNumeric(check.results[0])))

    value = check.value
    if check.mask is not None:
        value &= check.mask
        value += check.mask << 32
        assert check.varsize == 8
    else:
        assert check.varsize <= 4
    actions.append(action7.SkipAction(9, check.varnum, check.varsize, check.op, value, 1))

    actions.extend(parse_actionD(ParameterAssignment(assignment.param, expression.ConstantNumeric(check.results[1]))))
    return actions


def parse_grm(assignment):
    assert isinstance(assignment.value, expression.GRMOp)

    action6.free_parameters.save()
    action_list = []
    act6 = action6.Action6()
    assert isinstance(assignment.param, expression.Parameter)
    target = assignment.param.num
    if isinstance(target, expression.Parameter) and isinstance(target.num, expression.ConstantNumeric):
        act6.modify_bytes(target.num.value, 1, 1)
        target = expression.ConstantNumeric(0)
    elif not isinstance(target, expression.ConstantNumeric):
        tmp_param, tmp_param_actions = get_tmp_parameter(target)
        act6.modify_bytes(tmp_param, 1, 1)
        target = expression.ConstantNumeric(0)
        action_list.extend(tmp_param_actions)

    op = nmlop.ASSIGN
    param1 = assignment.value.op
    param2 = expression.ConstantNumeric(0xFE)
    data = expression.ConstantNumeric(0xFF | (assignment.value.feature << 8) | (assignment.value.count << 16))

    if len(act6.modifications) > 0:
        action_list.append(act6)

    action_list.append(ActionD(target, param1, op, param2, data))
    action6.free_parameters.restore()
    return action_list


def parse_hasbit(assignment):
    assert isinstance(assignment.value, expression.BinOp) and (
        assignment.value.op == nmlop.HASBIT or assignment.value.op == nmlop.NOTHASBIT
    )
    actions = parse_actionD(ParameterAssignment(assignment.param, expression.ConstantNumeric(0)))
    cond_block = conditional.Conditional(
        assignment.value, [ParameterAssignment(assignment.param, expression.ConstantNumeric(1))], None
    )
    actions.extend(conditional.ConditionalList([cond_block]).get_action_list())
    return actions


def parse_min_max(assignment):
    assert isinstance(assignment.value, expression.BinOp) and assignment.value.op in (nmlop.MIN, nmlop.MAX)
    # min(a, b) ==> a < b ? a : b.
    # max(a, b) ==> a > b ? a : b.
    action6.free_parameters.save()
    action_list = []
    expr1 = parse_subexpression(assignment.value.expr1, action_list)
    expr2 = parse_subexpression(assignment.value.expr2, action_list)
    guard = expression.BinOp(nmlop.CMP_LT if assignment.value.op == nmlop.MIN else nmlop.CMP_GT, expr1, expr2)
    action_list.extend(
        parse_actionD(ParameterAssignment(assignment.param, expression.TernaryOp(guard, expr1, expr2, None)))
    )
    action6.free_parameters.restore()
    return action_list


def parse_boolean(assignment):
    assert isinstance(assignment.value, expression.Boolean)
    actions = parse_actionD(ParameterAssignment(assignment.param, expression.ConstantNumeric(0)))
    expr = nmlop.CMP_NEQ(assignment.value.expr, 0)
    cond_block = conditional.Conditional(
        expr, [ParameterAssignment(assignment.param, expression.ConstantNumeric(1))], None
    )
    actions.extend(conditional.ConditionalList([cond_block]).get_action_list())
    return actions


def transform_bin_op(assignment):
    op = assignment.value.op
    expr1 = assignment.value.expr1
    expr2 = assignment.value.expr2
    extra_actions = []

    if op == nmlop.CMP_GE:
        expr1, expr2 = expr2, expr1
        op = nmlop.CMP_LE

    if op == nmlop.CMP_LE:
        extra_actions.extend(parse_actionD(ParameterAssignment(assignment.param, nmlop.SUB(expr1, expr2))))
        op = nmlop.CMP_LT
        expr1 = assignment.param
        expr2 = expression.ConstantNumeric(1)

    if op == nmlop.CMP_GT:
        expr1, expr2 = expr2, expr1
        op = nmlop.CMP_LT

    if op == nmlop.CMP_LT:
        extra_actions.extend(parse_actionD(ParameterAssignment(assignment.param, nmlop.SUB(expr1, expr2))))
        op = nmlop.SHIFTU_LEFT  # shift left by negative number = shift right
        expr1 = assignment.param
        expr2 = expression.ConstantNumeric(-31)

    elif op == nmlop.CMP_NEQ:
        extra_actions.extend(parse_actionD(ParameterAssignment(assignment.param, nmlop.SUB(expr1, expr2))))
        op = nmlop.DIV
        # We rely here on the (undocumented) behavior of both OpenTTD and TTDPatch
        # that expr/0==expr. What we do is compute A/A, which will result in 1 if
        # A != 0 and in 0 if A == 0
        expr1 = assignment.param
        expr2 = assignment.param

    elif op == nmlop.CMP_EQ:
        # We compute A==B by doing not(A - B) which will result in a value != 0
        # if A is equal to B
        extra_actions.extend(parse_actionD(ParameterAssignment(assignment.param, nmlop.SUB(expr1, expr2))))
        # Clamp the value to 0/1, see above for details
        extra_actions.extend(
            parse_actionD(ParameterAssignment(assignment.param, nmlop.DIV(assignment.param, assignment.param)))
        )
        op = nmlop.SUB
        expr1 = expression.ConstantNumeric(1)
        expr2 = assignment.param

    if op == nmlop.SHIFT_RIGHT or op == nmlop.SHIFTU_RIGHT:
        if isinstance(expr2, expression.ConstantNumeric):
            expr2.value *= -1
        else:
            expr2 = nmlop.SUB(0, expr2)
        op = nmlop.SHIFT_LEFT if op == nmlop.SHIFT_RIGHT else nmlop.SHIFTU_LEFT

    elif op == nmlop.XOR:
        # a ^ b ==> (a | b) - (a & b)
        expr1 = parse_subexpression(expr1, extra_actions)
        expr2 = parse_subexpression(expr2, extra_actions)
        tmp_param1, tmp_action_list1 = get_tmp_parameter(nmlop.OR(expr1, expr2))
        tmp_param2, tmp_action_list2 = get_tmp_parameter(nmlop.AND(expr1, expr2))
        extra_actions.extend(tmp_action_list1)
        extra_actions.extend(tmp_action_list2)
        expr1 = expression.Parameter(expression.ConstantNumeric(tmp_param1))
        expr2 = expression.Parameter(expression.ConstantNumeric(tmp_param2))
        op = nmlop.SUB

    return op, expr1, expr2, extra_actions


def parse_actionD(assignment):
    assignment.value.supported_by_actionD(True)

    if isinstance(assignment.param, expression.SpecialParameter):
        assignment.param, assignment.value = assignment.param.to_assignment(assignment.value)
    elif isinstance(assignment.param, expression.Identifier):
        if global_constants.identifier_refcount[assignment.param.value] == 0:
            # Named parameter is not referenced, ignoring
            return []
        assignment.param = expression.Parameter(
            expression.ConstantNumeric(global_constants.named_parameters[assignment.param.value]), assignment.param.pos
        )
    assert isinstance(assignment.param, expression.Parameter)

    if isinstance(assignment.value, expression.SpecialParameter):
        assignment.value = assignment.value.to_reading()

    if isinstance(assignment.value, expression.TernaryOp):
        return parse_ternary_op(assignment)

    if isinstance(assignment.value, expression.AbsOp):
        return parse_abs_op(assignment)

    if isinstance(assignment.value, expression.SpecialCheck):
        return parse_special_check(assignment)

    if isinstance(assignment.value, expression.GRMOp):
        return parse_grm(assignment)

    if isinstance(assignment.value, expression.BinOp):
        op = assignment.value.op
        if op == nmlop.HASBIT or op == nmlop.NOTHASBIT:
            return parse_hasbit(assignment)
        elif op == nmlop.MIN or op == nmlop.MAX:
            return parse_min_max(assignment)

    if isinstance(assignment.value, expression.Boolean):
        return parse_boolean(assignment)

    if isinstance(assignment.value, expression.Not):
        expr = nmlop.SUB(1, assignment.value.expr)
        assignment = ParameterAssignment(assignment.param, expr)

    if isinstance(assignment.value, expression.BinNot):
        expr = nmlop.SUB(0xFFFFFFFF, assignment.value.expr)
        assignment = ParameterAssignment(assignment.param, expr)

    action6.free_parameters.save()
    action_list = []
    act6 = action6.Action6()
    assert isinstance(assignment.param, expression.Parameter)
    target = assignment.param.num
    if isinstance(target, expression.Parameter) and isinstance(target.num, expression.ConstantNumeric):
        act6.modify_bytes(target.num.value, 1, 1)
        target = expression.ConstantNumeric(0)
    elif not isinstance(target, expression.ConstantNumeric):
        tmp_param, tmp_param_actions = get_tmp_parameter(target)
        act6.modify_bytes(tmp_param, 1, 1)
        target = expression.ConstantNumeric(0)
        action_list.extend(tmp_param_actions)

    data = None
    # print assignment.value
    if isinstance(assignment.value, expression.ConstantNumeric):
        op = nmlop.ASSIGN
        param1 = expression.ConstantNumeric(0xFF)
        param2 = expression.ConstantNumeric(0)
        data = assignment.value
    elif isinstance(assignment.value, expression.Parameter):
        if isinstance(assignment.value.num, expression.ConstantNumeric):
            op = nmlop.ASSIGN
            param1 = assignment.value.num
        else:
            tmp_param, tmp_param_actions = get_tmp_parameter(assignment.value.num)
            act6.modify_bytes(tmp_param, 1, 3)
            action_list.extend(tmp_param_actions)
            op = nmlop.ASSIGN
            param1 = expression.ConstantNumeric(0)
        param2 = expression.ConstantNumeric(0)
    elif isinstance(assignment.value, expression.OtherGRFParameter):
        op = nmlop.ASSIGN
        if isinstance(assignment.value.num, expression.ConstantNumeric):
            param1 = assignment.value.num
        else:
            tmp_param, tmp_param_actions = get_tmp_parameter(assignment.value.num)
            act6.modify_bytes(tmp_param, 1, 3)
            action_list.extend(tmp_param_actions)
            param1 = expression.ConstantNumeric(0)
        param2 = expression.ConstantNumeric(0xFE)
        data = expression.ConstantNumeric(expression.parse_string_to_dword(assignment.value.grfid))
    elif isinstance(assignment.value, expression.PatchVariable):
        op = nmlop.ASSIGN
        param1 = expression.ConstantNumeric(assignment.value.num)
        param2 = expression.ConstantNumeric(0xFE)
        data = expression.ConstantNumeric(0xFFFF)
    elif isinstance(assignment.value, expression.BinOp):
        op, expr1, expr2, extra_actions = transform_bin_op(assignment)
        action_list.extend(extra_actions)

        if isinstance(expr1, expression.ConstantNumeric):
            param1 = expression.ConstantNumeric(0xFF)
            data = expr1
        elif isinstance(expr1, expression.Parameter) and isinstance(expr1.num, expression.ConstantNumeric):
            param1 = expr1.num
        else:
            tmp_param, tmp_param_actions = get_tmp_parameter(expr1)
            action_list.extend(tmp_param_actions)
            param1 = expression.ConstantNumeric(tmp_param)

        # We can use the data only for one for the parameters.
        # If the first parameter uses "data" we need a temp parameter for this one
        if isinstance(expr2, expression.ConstantNumeric) and data is None:
            param2 = expression.ConstantNumeric(0xFF)
            data = expr2
        elif isinstance(expr2, expression.Parameter) and isinstance(expr2.num, expression.ConstantNumeric):
            param2 = expr2.num
        else:
            tmp_param, tmp_param_actions = get_tmp_parameter(expr2)
            action_list.extend(tmp_param_actions)
            param2 = expression.ConstantNumeric(tmp_param)

    else:
        raise generic.ScriptError("Invalid expression in argument assignment", assignment.value.pos)

    if len(act6.modifications) > 0:
        action_list.append(act6)

    action_list.append(ActionD(target, param1, op, param2, data))
    action6.free_parameters.restore()
    return action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic
from nml.ast import base_statement

feature_ids = {
    "FEAT_TRAINS": 0x00,
    "FEAT_ROADVEHS": 0x01,
    "FEAT_SHIPS": 0x02,
    "FEAT_AIRCRAFT": 0x03,
    "FEAT_STATIONS": 0x04,
    "FEAT_CANALS": 0x05,
    "FEAT_BRIDGES": 0x06,
    "FEAT_HOUSES": 0x07,
    "FEAT_GLOBALVARS": 0x08,
    "FEAT_INDUSTRYTILES": 0x09,
    "FEAT_INDUSTRIES": 0x0A,
    "FEAT_CARGOS": 0x0B,
    "FEAT_SOUNDEFFECTS": 0x0C,
    "FEAT_AIRPORTS": 0x0D,
    "FEAT_SIGNALS": 0x0E,
    "FEAT_OBJECTS": 0x0F,
    "FEAT_RAILTYPES": 0x10,
    "FEAT_AIRPORTTILES": 0x11,
    "FEAT_ROADTYPES": 0x12,
    "FEAT_TRAMTYPES": 0x13,
    "FEAT_ROADSTOPS": 0x14,
}


def feature_name(feature):
    """
    Given a feature number, return the identifier as normally used for that
    feature in nml files.

    @param feature: The feature number to convert to a string.
    @type feature: C{int}

    @return: String with feature as used in nml files.
    @rtype: C{str}
    """
    for name, num in feature_ids.items():
        if num == feature:
            return name
    raise AssertionError("Invalid feature number '{}'.".format(feature))


def parse_feature(expr):
    """
    Parse an expression into a valid feature number.

    @param expr: Expression to parse
    @type expr: L{Expression}

    @return: A constant number representing the parsed feature
    @rtype: L{ConstantNumeric}
    """
    expr = expr.reduce_constant([feature_ids])
    if expr.value not in feature_ids.values():
        raise generic.ScriptError("Invalid feature '{:02X}' encountered.".format(expr.value), expr.pos)
    return expr


class MainScript(base_statement.BaseStatementList):
    def __init__(self, statements):
        assert len(statements) > 0
        base_statement.BaseStatementList.__init__(
            self,
            "main script",
            statements[0].pos,
            base_statement.BaseStatementList.LIST_TYPE_NONE,
            statements,
            False,
            False,
            False,
            False,
        )

    def __str__(self):
        res = ""
        for stmt in self.statements:
            res += str(stmt) + "\n"
        return res



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, nmlop

from .base_expression import ConstantNumeric, Expression, Type


class BitMask(Expression):
    def __init__(self, values, pos):
        Expression.__init__(self, pos)
        self.values = values

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Get bitmask:")
        for value in self.values:
            value.debug_print(indentation + 2)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        ret = ConstantNumeric(0, self.pos)
        for orig_expr in self.values:
            val = orig_expr.reduce(id_dicts)
            if val.type() != Type.INTEGER:
                if val.type() == Type.SPRITEGROUP_REF:
                    raise generic.ProcCallSyntaxError(val.name, val.pos)
                raise generic.ScriptError("Parameters of 'bitmask' must be integers.", orig_expr.pos)
            if isinstance(val, ConstantNumeric) and val.value >= 32:
                raise generic.ScriptError("Parameters of 'bitmask' cannot be greater than 31", orig_expr.pos)
            val = nmlop.SHIFT_LEFT(1, val)
            ret = nmlop.OR(ret, val)
        return ret.reduce()

    def collect_references(self):
        from itertools import chain

        return list(chain.from_iterable(v.collect_references() for v in self.values))

    def is_read_only(self):
        return all(v.is_read_only() for v in self.values)

    def __str__(self):
        return "bitmask(" + ", ".join(str(e) for e in self.values) + ")"



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic

from .base_expression import Expression, Type


class FunctionPtr(Expression):
    """
    Pointer to a function.
    If this appears inside an expression, the user has made an error.

    @ivar name: Identifier that has been resolved to this function pointer.
    @type name: L{Identifier}

    @ivar func: Function that will be called to resolve this function call. Arguments:
                    Name of the function (C{str})
                    List of passed arguments (C{list} of L{Expression})
                    Position information (L{Position})
                    Any extra arguments passed to the constructor of this class
    @type func: C{function}

    @ivar extra_args List of arguments that should be passed to the function that is to be called.
    @type extra_args C{list}
    """

    def __init__(self, name, func, *extra_args):
        super().__init__(pos=None)
        self.name = name
        self.func = func
        self.extra_args = extra_args

    def debug_print(self, indentation):
        raise AssertionError("Function pointers should not appear inside expressions.")

    def __str__(self):
        raise AssertionError("Function pointers should not appear inside expressions.")

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        raise generic.ScriptError(
            "'{}' is a function and should be called using the function call syntax.".format(str(self.name)),
            self.name.pos,
        )

    def type(self):
        return Type.FUNCTION_PTR

    def call(self, args):
        return self.func(self.name.value, args, self.name.pos, *self.extra_args)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import datetime

from nml import expression, generic, nmlop
from nml.actions import action0
from nml.ast import base_statement


class Snowline(base_statement.BaseStatement):
    """
    Snowline curve throughout the year.

    @ivar type: Type of snowline.
    @type type: C{str}

    @ivar date_heights: Height of the snow line at given days in the year.
    @type date_heights: C{list} of L{UnitAssignment}
    """

    def __init__(self, line_type, height_data, pos):
        base_statement.BaseStatement.__init__(self, "snowline-block", pos)
        if line_type.value not in ("equal", "linear"):
            raise generic.ScriptError(
                'Unknown type of snow line (only "equal" and "linear" are supported)', line_type.pos
            )
        self.type = line_type.value
        self.date_heights = height_data

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Snowline (type={})".format(self.type))
        for dh in self.date_heights:
            dh.debug_print(indentation + 2)

    def __str__(self):
        return "snowline ({}) {{\n\t{}\n}}\n".format(str(self.type), "\n\t".join(str(x) for x in self.date_heights))

    def get_action_list(self):
        return action0.get_snowlinetable_action(compute_table(self))


def compute_table(snowline):
    """
    Compute the table with snowline height for each day of the year.

    @param snowline: Snowline definition.
    @type  snowline: L{Snowline}

    @return: Table of 12*32 entries with snowline heights.
    @rtype:  C{str}
    """
    day_table = [None] * 365  # Height at each day, starting at day 0
    for dh in snowline.date_heights:
        doy = dh.name.reduce()
        if not isinstance(doy, expression.ConstantNumeric):
            raise generic.ScriptError("Day of year is not a compile-time constant", doy.pos)
        if doy.value < 1 or doy.value > 365:
            raise generic.ScriptError("Day of the year must be between 1 and 365", doy.pos)

        height = dh.value.reduce()
        if isinstance(height, expression.ConstantNumeric) and height.value < 0:
            raise generic.ScriptError("Height must be at least 0", height.pos)
        if dh.unit is None:
            if isinstance(height, expression.ConstantNumeric) and height.value > 255:
                raise generic.ScriptError("Height must be at most 255", height.pos)
        else:
            unit = dh.unit
            if unit.type != "snowline":
                raise generic.ScriptError('Expected a snowline percentage ("snow%")', height.pos)

            if isinstance(height, expression.ConstantNumeric) and height.value > 100:
                raise generic.ScriptError("Height must be at most 100 snow%", height.pos)

            mul, div = unit.convert, 1
            if isinstance(mul, tuple):
                mul, div = mul

            # Factor out common factors
            gcd = generic.greatest_common_divisor(mul, div)
            mul //= gcd
            div //= gcd

            if isinstance(height, (expression.ConstantNumeric, expression.ConstantFloat)):
                # Even if mul == div == 1, we have to round floats and adjust value
                height = expression.ConstantNumeric(int(float(height.value) * mul / div + 0.5), height.pos)
            elif mul != div:
                # Compute (value * mul + div/2) / div
                height = nmlop.MUL(height, mul)
                height = nmlop.ADD(height, int(div / 2))
                height = nmlop.DIV(height, div)

        # For 'linear' snow-line, only accept integer constants.
        if snowline.type != "equal" and not isinstance(height, expression.ConstantNumeric):
            raise generic.ScriptError("Height is not a compile-time constant", height.pos)

        day_table[doy.value - 1] = height

    # Find first specified point.
    start = 0
    while start < 365 and day_table[start] is None:
        start = start + 1
    if start == 365:
        raise generic.ScriptError("No heights given for the snowline table", snowline.pos)

    first_point = start
    while True:
        # Find second point from start
        end = start + 1
        if end == 365:
            end = 0

        while end != first_point and day_table[end] is None:
            end = end + 1
            if end == 365:
                end = 0

        # Fill the days between start and end (exclusive both border values)
        startvalue = day_table[start]
        endvalue = day_table[end]
        unwrapped_end = end
        if end < start:
            unwrapped_end += 365

        if snowline.type == "equal":
            for day in range(start + 1, unwrapped_end):
                if day >= 365:
                    day -= 365
                day_table[day] = startvalue
        else:
            assert snowline.type == "linear"

            if start != end:
                dhd = float(endvalue.value - startvalue.value) / float(unwrapped_end - start)
            else:
                assert startvalue.value == endvalue.value
                dhd = 0

            for day in range(start + 1, unwrapped_end):
                uday = day
                if uday >= 365:
                    uday -= 365
                height = startvalue.value + int(round(dhd * (day - start)))
                day_table[uday] = expression.ConstantNumeric(height)

        if end == first_point:  # All days done
            break

        start = end

    table = [None] * (12 * 32)
    for dy in range(365):
        today = datetime.date.fromordinal(dy + 1)
        if day_table[dy]:
            expr = day_table[dy].reduce()
        else:
            expr = None
        table[(today.month - 1) * 32 + today.day - 1] = expr

    for idx, d in enumerate(table):
        if d is None:
            table[idx] = table[idx - 1]
    # Second loop is needed because we need make sure the first item is also set.
    for idx, d in enumerate(table):
        if d is None:
            table[idx] = table[idx - 1]

    return table



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, global_constants
from nml.actions import action7
from nml.ast import base_statement


class Loop(base_statement.BaseStatementList):
    """
    AST node for a while-loop.

    @ivar expr: The conditional to check whether the loop continues.
    @type expr: L{Expression}
    """

    def __init__(self, expr, block, pos):
        base_statement.BaseStatementList.__init__(
            self, "while-loop", pos, base_statement.BaseStatementList.LIST_TYPE_LOOP, block, in_item=True
        )
        self.expr = expr

    def pre_process(self):
        self.expr = self.expr.reduce(global_constants.const_list)
        base_statement.BaseStatementList.pre_process(self)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "While loop")
        generic.print_dbg(indentation + 2, "Expression:")
        self.expr.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Block:")
        base_statement.BaseStatementList.debug_print(self, indentation + 4)

    def get_action_list(self):
        return action7.parse_loop_block(self)

    def __str__(self):
        ret = "while({}) {{\n".format(self.expr)
        ret += base_statement.BaseStatementList.__str__(self)
        ret += "}\n"
        return ret



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import itertools

from nml import generic, nmlop, global_constants
from nml.expression import (
    AcceptCargo,
    Array,
    BitMask,
    ConstantFloat,
    ConstantNumeric,
    Identifier,
    ProduceCargo,
    StringLiteral,
    parse_string_to_dword,
)

tilelayout_names = {}


class BaseAction0Property:
    """
    Base class for Action0 properties.
    """

    def write(self, file):
        """
        Write this property to the output given by file.

        @param file: The outputfile we have to write to.
        @type  file: L{SpriteOutputBase}
        """
        raise NotImplementedError("write is not implemented in {!r}".format(type(self)))

    def get_size(self):
        """
        Get the number of bytes that this property will write to
        the output.

        @return: The size of this property in bytes.
        @rtype:  C{int}
        """
        raise NotImplementedError("get_size is not implemented in {!r}".format(type(self)))


class Action0Property(BaseAction0Property):
    """
    Simple Action 0 property with a fixed size.

    @ivar num: Number of the property.
    @type num: C{int}

    @ivar values: Value of the property for each id.
    @type values: C{list} of L{ConstantNumeric}

    @ivar size: Size of the storage, in bytes.
    @type size: C{int}
    """

    def __init__(self, num, value, size):
        self.num = num
        self.values = value if isinstance(value, list) else [value]
        self.size = size

        # Make sure the value fits in the size.
        # Strings have their own check in parse_property
        for val in self.values:
            if not isinstance(val, StringLiteral):
                biggest = 1 << (8 * size)
                if val.value >= biggest:
                    raise generic.ScriptError("Action 0 property too large", val.pos)
                elif val.value < 0 and val.value + (biggest // 2) < 0:
                    raise generic.ScriptError("Action 0 property too small", val.pos)

    def write(self, file):
        file.print_bytex(self.num)
        for val in self.values:
            val.write(file, self.size)
        file.newline()

    def get_size(self):
        return self.size * len(self.values) + 1


# @var properties: A mapping of features to properties. This is a list
# with one item per feature. Entries should be a dictionary of properties,
# or C{None} if no properties are defined for that feature.
#
# Each property is a mapping of property name to its characteristics.
# These characteristics are either a dictionary or a list of
# dictionaries, the latter can be used to set multiple properties.
# First a short summary is given, then the recognized characteristics
# are outlined below in more detail.
#
# Summary: If 'string' or 'string_literal' is set, the value should be a
# string or literal string, else the value is a number. 'unit_type' and
# 'unit_conversion' are used to convert user-entered values to nfo values.
# If some more arithmetic is needed to convert the entered value into a nfo
# value, 'value_function' can be used. For even more complicated things,
# 'custom_function' can be used to create a special mapping of the value to nfo
# properties, else 'num' and 'size' are used to provide a 'normal' action0.
#
# 'string', if set, means that the value of the property should be a string.
# The value of characteristic indicates the string range to use (usually 0xD0 or 0xDC)
# If set to None, the string will use the ID of the item (used for vehicle names)
# It can also be a (threshold, below, above) tuple to select range depending on ID
#
# 'string_literal', if set, indicates that the value of the property should
# be a literal (quoted) string. The value of the characteristic is equal to
# the required length (usually 4) of said literal string.
#
# 'unit_type' means that units of the given type (power, speed) can be applied
# to this property. A list of units can be found in ../unit.py. The value is then
# converted to a certain reference unit (for example m/s for speed)
# Leaving this unset means that no units (except 'nfo' which is an identity mapping)
# can be applied to this property.
#
# 'unit_conversion' defines a conversion factor between the value entered by
# the user and the resulting value in nfo. This is either an integer or a
# rational number entered as a 2-tuple (numerator, denominator).
# The entered value (possibly converted # to the appropriate reference unit,
# see 'unit_type' above) is multiplied by this factor and then rounded to an
# integer to provide the final value.
# This parameter is not required and defaults to 1.
#
# 'value_function' can be used to alter the mapping of nml values to nfo values
# it takes one argument (the value) and returns the new value to use
# Both the parameter and the return value are expressions that do not have to be
# constants
#
# 'custom_function' can be used to bypass the normal way of converting the
# name / value to an Action0Property. This function is normally called with one
# argument, which is the value of the property. For houses, there may be multiple
# values, passed as a vararg-list. It should return a list of Action0Property.
# To pass extra parameters to the function, a dose of lambda calculus can be used.
# Consult the code for examples.
#
# 'test_function' can be used to determine if the property should be set in the
# first place. It normally takes one argument (the value) and should return True
# if the property is to be set, False if it is to be ignored. Default is True
# For houses, there may be multiple values, passed as a vararg-list.
#
# 'num' is the Action0 property number of the action 0 property, as given by the
# nfo specs. If set to -1, no action0 property will be generated. If
# 'custom_function' is set, this value is not needed and can be left out.
# It can also be a (threshold, below, above) tuple to set 'num' depending on ID.
#
# 'size' is the size (in bytes) of the resulting action 0 property. Valid
# values are 1 (byte), 2 (word) or 4 (dword). For other (or variable) sizes,
# 'custom_function' is needed. If 'custom_function' is set or 'num' is equal
# to -1, this parameter is not needed and can be left out.
#
# 'warning' is a string (optional) containing a warning message that will be
# shown if a property is used. Use for deprecating properties.
#
# 'first' (value doesn't matter) if the property should be set first (generally a substitute type).
#
# 'required' (value doesn't matter) if the property is required for the item to be valid.

properties = 0x15 * [None]

#
# Some helper functions that are used for multiple features
#


def two_byte_property(low_prop, high_prop, low_prop_info=None, high_prop_info=None):
    """
    Decode a two byte value into two action 0 properties.

    @param low_prop: Property number for the low 8 bits of the value.
    @type  low_prop: C{int}

    @param high_prop: Property number for the high 8 bits of the value.
    @type  high_prop: C{int}

    @param low_prop_info: Dictionary with additional property information for the low byte.
    @type low_prop_info: C{dict}

    @param high_prop_info: Dictionary with additional property information for the low byte.
    @type high_prop_info: C{dict}

    @return: Sequence of two dictionaries with property information (low part, high part).
    @rtype:  C{list} of C{dict}
    """
    low_byte_info = {
        **(low_prop_info or {}),
        "num": low_prop,
        "size": 1,
        "value_function": lambda value: nmlop.AND(value, 0xFF).reduce(),
    }
    high_byte_info = {
        **(high_prop_info or {}),
        "num": high_prop,
        "size": 1,
        "value_function": lambda value: nmlop.SHIFT_RIGHT(value, 8).reduce(),
    }
    return [low_byte_info, high_byte_info]


def animation_info(value, loop_bit=8, max_frame=253):
    """
    Convert animation info array of two elements to an animation info property.
    The first is 0/1, and defines whether the animation loops.
    The second is the number of frames, at most 253 frames.

    @param value: Array of animation info.
    @type  value: C{Array}

    @param loop_bit: Bit the loop information is stored.
    @type  loop_bit: C{int}

    @param max_frame: Max frames possible.
    @type  max_frame: C{int}

    @return: Value to use for animation property.
    @rtype:  L{Expression}
    """
    if not isinstance(value, Array) or len(value.values) != 2:
        raise generic.ScriptError("animation_info must be an array with exactly 2 constant values", value.pos)
    looping = value.values[0].reduce_constant().value
    frames = value.values[1].reduce_constant().value
    if looping not in (0, 1):
        raise generic.ScriptError("First field of the animation_info array must be either 0 or 1", value.values[0].pos)
    if frames < 1 or frames > max_frame:
        raise generic.ScriptError(
            "Second field of the animation_info array must be between 1 and " + str(max_frame), value.values[1].pos
        )

    return ConstantNumeric((looping << loop_bit) + frames - 1)


def cargo_list(value, max_num_cargos):
    """
    Encode an array of cargo types in a single property value. If less than the maximum
    number of cargos are given the rest is filled up with 0xFF (=invalid cargo).

    @param value: Array of cargo types.
    @type  value: C{Array}

    @param max_num_cargos: The maximum number of cargos in the array.
    @type  max_num_cargos: C{int}

    @param prop_num: Property number.
    @type  prop_num: C{int}

    @param prop_size: Property size in bytes.
    @type  prop_size: C{int}
    """
    if not isinstance(value, Array) or len(value.values) > max_num_cargos:
        raise generic.ScriptError(
            "Cargo list must be an array with no more than {:d} values".format(max_num_cargos), value.pos
        )
    cargoes = value.values + [ConstantNumeric(0xFF, value.pos) for _ in range(max_num_cargos - len(value.values))]

    ret = None
    for i, cargo in enumerate(cargoes):
        byte = nmlop.AND(cargo, 0xFF)
        if i == 0:
            ret = byte
        else:
            byte = nmlop.SHIFT_LEFT(byte, i * 8)
            ret = nmlop.OR(ret, byte)
    return ret.reduce()


#
# General vehicle properties that apply to feature 0x00 .. 0x03
#

# fmt: off
general_veh_props = {
    "reliability_decay":  {"size": 1, "num": 0x02},
    "vehicle_life":       {"size": 1, "num": 0x03},
    "model_life":         {"size": 1, "num": 0x04},
    "climates_available": {"size": 1, "num": 0x06},
    "loading_speed":      {"size": 1, "num": 0x07},
    "name":               {"num": -1, "string": None},
}
# fmt: on


def ottd_display_speed(value, mul, div, unit):
    # Convert value to km/h-ish.
    kmh_ish = value.value * mul // div
    # Duplicate OpenTTD's `ConvertKmhishSpeedToDisplaySpeed()`
    return ((10 * kmh_ish * unit.ottd_mul) >> unit.ottd_shift) // 16


class VariableListProp(BaseAction0Property):
    """
    Property value that is a variable-length list of variable sized values, the list length is written before the data.
    """

    def __init__(self, prop_num, data, size, extended):
        # data is a list, each element belongs to an item ID
        # Each element in the list is a list of cargo types
        self.prop_num = prop_num
        self.data = data
        self.size = size
        self.extended = extended

    def write(self, file):
        file.print_bytex(self.prop_num)
        for elem in self.data:
            if self.extended:
                file.print_bytex(0xFF)
                file.print_word(len(elem))
            else:
                file.print_byte(len(elem))
            for i, val in enumerate(elem):
                if i % 8 == 0:
                    file.newline()
                file.print_varx(val, self.size)
            file.newline()

    def get_size(self):
        total_len = 1  # Prop number
        for elem in self.data:
            # For each item ID to set, make space for all values + 3 or 1 for the length
            total_len += len(elem) * self.size + (3 if self.extended else 1)
        return total_len


def VariableByteListProp(prop_num, data, extended=False):
    return VariableListProp(prop_num, data, 1, extended)


def ctt_list(prop_num, *values):
    # values may have multiple entries, if more than one item ID is set (e.g. multitile houses)
    # Each value is an expression.Array of cargo types
    for value in values:
        if not isinstance(value, Array):
            raise generic.ScriptError("Value of cargolist property must be an array", value.pos)
    return [
        VariableByteListProp(
            prop_num,
            [[ctype.reduce_constant().value for ctype in single_item_array.values] for single_item_array in values],
        )
    ]


def VariableWordListProp(num_prop, data, extended=False):
    return VariableListProp(num_prop, data, 2, extended)


def accepted_cargos(prop_num, *values):
    # values may have multiple entries, if more than one item ID is set (e.g. multitile houses)
    # Each value is an expression.Array of cargo types and amount arrays
    cargos = []
    for value in values:
        if not isinstance(value, Array) or len(value.values) > 16:
            raise generic.ScriptError("accepted_cargos must be an array with no more than 16 values", value.pos)
        tile_cargos = []
        for cargo_amount_pair in value.values:
            if not isinstance(cargo_amount_pair, Array) or len(cargo_amount_pair.values) != 2:
                raise generic.ScriptError(
                    "Each element of accepted_cargos must be an array with two elements: cargoid and amount",
                    cargo_amount_pair.pos,
                )
            cargo_id = cargo_amount_pair.values[0].reduce_constant().value
            cargo_amount = cargo_amount_pair.values[1].reduce_constant().value
            tile_cargos.append((cargo_amount << 8) | cargo_id)
        cargos.append(tile_cargos)
    return [VariableWordListProp(prop_num, cargos)]


def vehicle_length(value):
    if isinstance(value, ConstantNumeric):
        generic.check_range(value.value, 1, 8, "vehicle length", value.pos)
    return nmlop.SUB(8, value).reduce()


def zero_refit_mask(prop_num):
    # Zero the refit mask, in addition to setting some other refit property
    return {"size": 4, "num": prop_num, "value_function": lambda value: ConstantNumeric(0)}


def refittable_cargo_classes(prop_or_num, prop_and_num, refit_mask_num):
    def prop_value(value, index):
        if isinstance(value, ConstantNumeric):
            return value if index == 0 else None
        if isinstance(value, Array) and len(value.values) == 2 and isinstance(value.values[index], ConstantNumeric):
            return value.values[index]
        raise generic.ScriptError("refittable_cargo_classes must be a constant or an array of 2 constants")

    def prop_test(value):
        return value is not None

    return [
        {"size": 2, "num": prop_or_num, "value_function": lambda value: prop_value(value, 0)},
        {
            "size": 2,
            "num": prop_and_num,
            "value_function": lambda value: prop_value(value, 1),
            "test_function": prop_test,
        },
        zero_refit_mask(refit_mask_num),
    ]


def single_or_list(prop_name, single_num, multiple_num, value):
    if isinstance(value, ConstantNumeric):
        return [Action0Property(single_num, value, 1)]
    if isinstance(value, Array):
        if len(value.values) == 1:
            return [Action0Property(single_num, value.values[0].reduce_constant(), 1)]
        return [VariableByteListProp(multiple_num, [[type.reduce_constant().value for type in value.values]])]
    raise generic.ScriptError("'{}' must be a constant or an array of constants".format(prop_name))


#
# Feature 0x00 (Trains)
#

# fmt: off
properties[0x00] = {
    **general_veh_props,
    "track_type": {
        "custom_function": lambda value: single_or_list("track_type", 0x05, 0x34, value)
    },
    "ai_special_flag":                {"size": 1, "num": 0x08},
    "speed": {
        "size": 2,
        "num": 0x09,
        "unit_type": "speed",
        "unit_conversion": (5000, 1397),
        "adjust_value": lambda val, unit: ottd_display_speed(val, 1, 1, unit)
    },
    # 09 doesn't exist
    "power":                          {"size": 2, "num": 0x0B, "unit_type": "power"},
    # 0A doesn't exist
    "running_cost_factor":            {"size": 1, "num": 0x0D},
    "running_cost_base":              {"size": 4, "num": 0x0E},
    # 0F -11 don't exist
    "sprite_id":                      {"size": 1, "num": 0x12},
    "dual_headed":                    {"size": 1, "num": 0x13},
    "cargo_capacity":                 {"size": 1, "num": 0x14},
    "default_cargo_type":             {"size": 1, "num": 0x15},
    "weight": two_byte_property(
        0x16,
        0x24,
        {"unit_type": "weight"},
        {"unit_type": "weight"}
    ),
    "cost_factor":                    {"size": 1, "num": 0x17},
    "ai_engine_rank":                 {"size": 1, "num": 0x18},
    "engine_class":                   {"size": 1, "num": 0x19},
    # 1A (sort purchase list) is implemented elsewhere
    "extra_power_per_wagon":          {"size": 2, "num": 0x1B, "unit_type": "power"},
    "refit_cost":                     {"size": 1, "num": 0x1C},
    # 1D (refittable cargo types) is removed, it is zeroed when setting a different refit property
    # 1E (callback flags) is not set by user
    "tractive_effort_coefficient":    {"size": 1, "num": 0x1F, "unit_conversion": 255},
    "air_drag_coefficient":           {"size": 1, "num": 0x20, "unit_conversion": 255},
    "length":                         {"size": 1, "num": 0x21, "value_function": vehicle_length},
    # 22 has two names, to simplify docs
    "visual_effect_and_powered":      {"size": 1, "num": 0x22},
    "effect_spawn_model_and_powered": {"size": 1, "num": 0x22},
    "extra_weight_per_wagon":         {"size": 1, "num": 0x23, "unit_type": "weight"},
    # 24 is high byte of 16 (weight)
    "bitmask_vehicle_info":           {"size": 1, "num": 0x25},
    "retire_early":                   {"size": 1, "num": 0x26},
    "misc_flags":                     {"size": 1, "num": 0x27},
    "refittable_cargo_classes":       refittable_cargo_classes(0x28, 0x32, 0x1D),
    "non_refittable_cargo_classes":   [{"size": 2, "num": 0x29}, zero_refit_mask(0x1D)],
    "introduction_date":              {"size": 4, "num": 0x2A},
    "cargo_age_period":               {"size": 2, "num": 0x2B},
    "cargo_allow_refit": [
        {"custom_function": lambda value: ctt_list(0x2C, value)},
        zero_refit_mask(0x1D)
    ],
    "cargo_disallow_refit": [
        {"custom_function": lambda value: ctt_list(0x2D, value)}
        , zero_refit_mask(0x1D)
    ],
    "curve_speed_mod":                {"size": 2, "num": 0x2E, "unit_conversion": 256},
    "variant_group":                  {"size": 2, "num": 0x2F},
    "extra_flags":                    {"size": 4, "num": 0x30},
}
# fmt: on

#
# Feature 0x01 (Road Vehicles)
#


def roadveh_speed_prop(prop_info):
    # prop 08 value is min(value, 255)
    def prop08_value(value):
        return nmlop.MIN(value, 0xFF).reduce()

    # prop 15 value is (value + 3) / 4
    def prop15_value(value):
        return nmlop.DIV(nmlop.ADD(value, 3), 4).reduce()

    # prop 15 should not be set if value(prop08_value) <= 255.
    # But as we test prop15 and prop15 = 0.25/prop08, test for 64:
    def prop15_test(value):
        return isinstance(value, ConstantNumeric) and value.value >= 0x40

    prop08 = {"size": 1, "num": 0x08, "value_function": prop08_value}
    prop15 = {"size": 1, "num": 0x15, "value_function": prop15_value, "test_function": prop15_test}
    for key in prop_info:
        prop08[key] = prop15[key] = prop_info[key]
    return [prop08, prop15]


# fmt: off
properties[0x01] = {
    **general_veh_props,
    "road_type":                    {"size": 1, "num": 0x05},
    "tram_type":                    {"size": 1, "num": 0x05},
    "speed": roadveh_speed_prop(
        {
            "unit_type": "speed",
            "unit_conversion": (10000, 1397),
            "adjust_value": lambda val, unit: ottd_display_speed(val, 1, 2, unit),
        }
    ),
    "running_cost_factor":          {"size": 1, "num": 0x09},
    "running_cost_base":            {"size": 4, "num": 0x0A},
    # 0B -0D don"t exist
    "sprite_id":                    {"size": 1, "num": 0x0E},
    "cargo_capacity":               {"size": 1, "num": 0x0F},
    "default_cargo_type":           {"size": 1, "num": 0x10},
    "cost_factor":                  {"size": 1, "num": 0x11},
    "sound_effect":                 {"size": 1, "num": 0x12},
    "power":                        {"size": 1, "num": 0x13, "unit_type": "power", "unit_conversion": (1, 10)},
    "weight":                       {"size": 1, "num": 0x14, "unit_type": "weight", "unit_conversion": 4},
    # 15 is set together with 08 (see above)
    # 16 (refittable cargo types) is removed, it is zeroed when setting a different refit property
    # 17 (callback flags) is not set by user
    "tractive_effort_coefficient":  {"size": 1, "num": 0x18, "unit_conversion": 255},
    "air_drag_coefficient":         {"size": 1, "num": 0x19, "unit_conversion": 255},
    "refit_cost":                   {"size": 1, "num": 0x1A},
    "retire_early":                 {"size": 1, "num": 0x1B},
    "misc_flags":                   {"size": 1, "num": 0x1C},
    "refittable_cargo_classes":     refittable_cargo_classes(0x1D, 0x29, 0x16),
    "non_refittable_cargo_classes": [{"size": 2, "num": 0x1E}, zero_refit_mask(0x16)],
    "introduction_date":            {"size": 4, "num": 0x1F},
    # 20 (sort purchase list) is implemented elsewhere
    # 21 has two names, to simplify docs
    "visual_effect":                {"size": 1, "num": 0x21},
    "effect_spawn_model":           {"size": 1, "num": 0x21},
    "cargo_age_period":             {"size": 2, "num": 0x22},
    "length":                       {"size": 1, "num": 0x23, "value_function": vehicle_length},
    "cargo_allow_refit": [
        {"custom_function": lambda value: ctt_list(0x24, value)},
        zero_refit_mask(0x16),
    ],
    "cargo_disallow_refit": [
        {"custom_function": lambda value: ctt_list(0x25, value)},
        zero_refit_mask(0x16),
    ],
    "variant_group":                {"size": 2, "num": 0x26},
    "extra_flags":                  {"size": 4, "num": 0x27},
}
# fmt: on

#
# Feature 0x02 (Ships)
#


def speed_fraction(value):
    # Unit is already converted to 0 .. 255 range when we get here
    if isinstance(value, ConstantNumeric) and not (0 <= value.value <= 255):
        # Do not use check_range to provide better error message
        raise generic.ScriptError("speed fraction must be in range 0 .. 1", value.pos)
    return nmlop.SUB(255, value).reduce()


def ship_speed_prop(prop_info):
    # prop 0B value is min(value, 255)
    def prop0B_value(value):
        return nmlop.MIN(value, 0xFF).reduce()

    # prop 23 value is min(value, 65535)
    def prop23_value(value):
        return nmlop.MIN(value, 0xFFFF).reduce()

    # prop 23 should not be set if value(prop0B_value) <= 255.
    def prop23_test(value):
        return isinstance(value, ConstantNumeric) and value.value >= 0xFF

    prop0B = {"size": 1, "num": 0x0B, "value_function": prop0B_value}
    prop23 = {"size": 2, "num": 0x23, "value_function": prop23_value, "test_function": prop23_test}
    for key in prop_info:
        prop0B[key] = prop23[key] = prop_info[key]
    return [prop0B, prop23]


# fmt: off
properties[0x02] = {
    **general_veh_props,
    "sprite_id":                    {"size": 1, "num": 0x08},
    "is_refittable":                {"size": 1, "num": 0x09},
    "cost_factor":                  {"size": 1, "num": 0x0A},
    "speed": ship_speed_prop(
        {
            "unit_type": "speed",
            "unit_conversion": (10000, 1397),
            "adjust_value": lambda val, unit: ottd_display_speed(val, 1, 2, unit),
        }
    ),
    "default_cargo_type":           {"size": 1, "num": 0x0C},
    "cargo_capacity":               {"size": 2, "num": 0x0D},
    # 0E does not exist
    "running_cost_factor":          {"size": 1, "num": 0x0F},
    "sound_effect":                 {"size": 1, "num": 0x10},
    # 11 (refittable cargo types) is removed, it is zeroed when setting a different refit property
    # 12 (callback flags) is not set by user
    "refit_cost":                   {"size": 1, "num": 0x13},
    "ocean_speed_fraction": {
        "size": 1,
        "num": 0x14,
        "unit_conversion": 255,
        "value_function": speed_fraction,
    },
    "canal_speed_fraction": {
        "size": 1,
        "num": 0x15,
        "unit_conversion": 255,
        "value_function": speed_fraction,
    },
    "retire_early":                 {"size": 1, "num": 0x16},
    "misc_flags":                   {"size": 1, "num": 0x17},
    "refittable_cargo_classes":     refittable_cargo_classes(0x18, 0x25, 0x11),
    "non_refittable_cargo_classes": [{"size": 2, "num": 0x19}, zero_refit_mask(0x11)],
    "introduction_date":            {"size": 4, "num": 0x1A},
    # 1B (sort purchase list) is implemented elsewhere
    # 1C has two names, to simplify docs
    "visual_effect":                {"size": 1, "num": 0x1C},
    "effect_spawn_model":           {"size": 1, "num": 0x1C},
    "cargo_age_period":             {"size": 2, "num": 0x1D},
    "cargo_allow_refit": [
        {"custom_function": lambda value: ctt_list(0x1E, value)},
        zero_refit_mask(0x11),
    ],
    "cargo_disallow_refit": [
        {"custom_function": lambda value: ctt_list(0x1F, value)},
        zero_refit_mask(0x11),
    ],
    "variant_group":                {"size": 2, "num": 0x20},
    "extra_flags":                  {"size": 4, "num": 0x21},
    "acceleration":                 {"size": 1, "num": 0x24},
}
# fmt: on

#
# Feature 0x03 (Aircraft)
#


def aircraft_is_heli(value):
    if isinstance(value, ConstantNumeric) and value.value not in (0, 2, 3):
        raise generic.ScriptError("Invalid value for aircraft_type", value.pos)
    return nmlop.AND(value, 2).reduce()


def aircraft_is_large(value):
    return nmlop.AND(value, 1).reduce()


# fmt: off
properties[0x03] = {
    **general_veh_props,
    "sprite_id":                    {"size": 1, "num": 0x08},
    "aircraft_type": [
        {"size": 1, "num": 0x09, "value_function": aircraft_is_heli},
        {"size": 1, "num": 0x0A, "value_function": aircraft_is_large}
    ],
    "cost_factor":                  {"size": 1, "num": 0x0B},
    "speed": {
        "size": 1,
        "num": 0x0C,
        "unit_type": "speed",
        "unit_conversion": (701, 2507),
        "adjust_value": lambda val, unit: ottd_display_speed(val, 128, 10, unit)
    },
    "acceleration":                 {"size": 1, "num": 0x0D},
    "running_cost_factor":          {"size": 1, "num": 0x0E},
    "passenger_capacity":           {"size": 2, "num": 0x0F},
    # 10 does not exist
    "mail_capacity":                {"size": 1, "num": 0x11},
    "sound_effect":                 {"size": 1, "num": 0x12},
    # 13 (refittable cargo types) is removed, it is zeroed when setting a different refit property
    # 14 (callback flags) is not set by user
    "refit_cost":                   {"size": 1, "num": 0x15},
    "retire_early":                 {"size": 1, "num": 0x16},
    "misc_flags":                   {"size": 1, "num": 0x17},
    "refittable_cargo_classes":     refittable_cargo_classes(0x18, 0x23, 0x13),
    "non_refittable_cargo_classes": [{"size": 2, "num": 0x19}, zero_refit_mask(0x13)],
    "introduction_date":            {"size": 4, "num": 0x1A},
    # 1B (sort purchase list) is implemented elsewhere
    "cargo_age_period":             {"size": 2, "num": 0x1C},
    "cargo_allow_refit": [
        {"custom_function": lambda value: ctt_list(0x1D, value)},
        zero_refit_mask(0x13),
    ],
    "cargo_disallow_refit": [
        {"custom_function": lambda value: ctt_list(0x1E, value)},
        zero_refit_mask(0x13),
    ],
    "range":                        {"size": 2, "num": 0x1F},
    "variant_group":                {"size": 2, "num": 0x20},
    "extra_flags":                  {"size": 4, "num": 0x21},
}
# fmt: on

#
# Feature 0x04 (Stations)
#


def station_platforms_length(value):
    # Writing bitmask(2) to disable platform/length 3 is not very intuitive.
    # Instead, we expect the user will write bitmask(3) and we shift the result.
    return nmlop.SHIFT_RIGHT(value, 1, value.pos).reduce()


def station_flags(value):
    # bit 4 (extended foundations) can't be set without bit 3 (custom foundations)
    cust_found = nmlop.SHIFT_RIGHT(value, 4, value.pos)
    cust_found = nmlop.AND(cust_found, 1)
    cust_found = nmlop.MUL(cust_found, 0x08)
    return nmlop.OR(value, cust_found).reduce()


def cargo_bitmask(value):
    if not isinstance(value, Array):
        raise generic.ScriptError("Cargo list must be an array", value.pos)
    return BitMask(value.values, value.pos).reduce()


class StationLayoutProp(BaseAction0Property):
    def __init__(self, prop_num, data):
        self.prop_num = prop_num
        self.data = data

    def write(self, file):
        file.print_bytex(self.prop_num)
        for (length, number), layout in self.data.items():
            file.print_byte(length)
            file.print_byte(number)
            file.newline()
            for platform in layout:
                for type in platform:
                    file.print_bytex(type)
                file.newline()
        file.print_byte(0)
        file.print_byte(0)
        file.newline()

    def get_size(self):
        total_len = 3  # Prop number + ending
        for length, number in self.data:
            total_len += length * number + 2
        return total_len


def station_layouts(value):
    if isinstance(value, ConstantNumeric):
        return [Action0Property(0x0F, value, 1 if value.value < 0xFF else 3)]
    if not isinstance(value, Array) or len(value.values) == 0:
        raise generic.ScriptError("station_layouts must be an array of layouts, or the ID of a station", value.pos)
    layouts = {}
    for layout in value.values:
        if not isinstance(layout, Array) or len(layout.values) == 0:
            raise generic.ScriptError("A station layout must be an array of platforms", layout.pos)
        length = len(layout.values[0].values)
        number = len(layout.values)
        if (length, number) in layouts:
            generic.print_warning(
                generic.Warning.GENERIC, "Redefinition of layout {}x{}".format(length, number), layout.pos
            )
        layouts[(length, number)] = []
        for platform in layout.values:
            if not isinstance(platform, Array) or len(platform.values) == 0:
                raise generic.ScriptError("A platform must be an array of tile types")
            if len(platform.values) != length:
                raise generic.ScriptError("All platforms in a station layout must have the same length", platform.pos)
            for type in platform.values:
                if type.reduce_constant().value % 2 != 0:
                    generic.print_warning(
                        generic.Warning.GENERIC,
                        "Invalid tile {} in layout {}x{}".format(type, length, number),
                        type.pos,
                    )
            layouts[(length, number)].append(
                [nmlop.AND(type, 0xFE).reduce_constant().value for type in platform.values]
            )
    return [StationLayoutProp(0x0E, layouts)]


def station_tile_flags(value):
    if not isinstance(value, Array) or len(value.values) % 2 != 0:
        raise generic.ScriptError("Flag list must be an array of even length", value.pos)
    if len(value.values) > 8:
        return [VariableByteListProp(0x1E, [[flags.reduce_constant().value for flags in value.values]], True)]
    pylons = 0
    wires = 0
    blocked = 0
    for i, val in enumerate(value.values):
        flag = val.value
        if flag & 1 << global_constants.constant_numbers["STAT_TILE_PYLON"]:
            pylons = pylons | 1 << i
        if flag & 1 << global_constants.constant_numbers["STAT_TILE_NOWIRE"]:
            wires = wires | 1 << i
        if flag & 1 << global_constants.constant_numbers["STAT_TILE_BLOCKED"]:
            blocked = blocked | 1 << i
    return [
        Action0Property(0x11, ConstantNumeric(pylons), 1),
        Action0Property(0x14, ConstantNumeric(wires), 1),
        Action0Property(0x15, ConstantNumeric(blocked), 1),
    ]


def station_tile_list(value, prop_num, description):
    if not isinstance(value, Array) or len(value.values) % 2 != 0:
        raise generic.ScriptError(f"{description} list must be an array of even length", value.pos)
    return [VariableByteListProp(prop_num, [[x.reduce_constant().value for x in value.values]], True)]


# fmt: off
properties[0x04] = {
    "class":                 {"size": 4, "num": 0x08, "first": None, "string_literal": 4},
    # 09 (sprite layout) is implemented elsewhere
    # 0A (copy sprite layout) is implemented elsewhere
    # 0B (callback flags) is not set by user
    "disabled_platforms":    {"size": 1, "num": 0x0C, "value_function": station_platforms_length},
    "disabled_length":       {"size": 1, "num": 0x0D, "value_function": station_platforms_length},
    "station_layouts":       {"custom_function": station_layouts},  # = prop 0E
    # 0F (copy station layout) is handled by station_layouts
    "cargo_threshold":       {"size": 2, "num": 0x10},
    "draw_pylon_tiles":      {"size": 1, "num": 0x11, "replaced_by": "tile_flags"},
    "cargo_random_triggers": {"size": 4, "num": 0x12, "value_function": cargo_bitmask},
    "general_flags":         {"size": 1, "num": 0x13, "value_function": station_flags},
    "hide_wire_tiles":       {"size": 1, "num": 0x14, "replaced_by": "tile_flags"},
    "non_traversable_tiles": {"size": 1, "num": 0x15, "replaced_by": "tile_flags"},
    "animation_info":        {"size": 2, "num": 0x16, "value_function": animation_info},
    "animation_speed":       {"size": 1, "num": 0x17},
    "animation_triggers":    {"size": 2, "num": 0x18},
    # 19 (road routing) reserved for future use
    # 1A (advanced sprite layout) is implemented elsewhere
    # 1B (minimum bridge height) JGR only
    "name":                  {"size": 2, "num": (256, -1, 0x1C), "string": (256, 0xC5, 0xDC), "required": True},
    "classname":             {"size": 2, "num": (256, -1, 0x1D), "string": (256, 0xC4, 0xDC)},
    "tile_flags":            {"custom_function": station_tile_flags},  # = prop 1E
    "heights":               {"custom_function": lambda x: station_tile_list(x, 0x20, "Station height")},
    "blocked_pillars":       {"custom_function": lambda x: station_tile_list(x, 0x21, "Station blocked pillar")},
}
# fmt: on

#
# Feature 0x05 (Canals)
#

properties[0x05] = {
    # 08 (callback flags) not set by user
    "graphic_flags": {"size": 1, "num": 0x09},
}

# TODO: Feature 0x06

#
# Feature 0x07 (Houses)
#


def house_prop_0A(value):
    # User sets an array [min_year, max_year] as property value
    # House property 0A is set to ((max_year - 1920) << 8) | (min_year - 1920)
    # With both bytes clamped to the 0 .. 255 range
    if not isinstance(value, Array) or len(value.values) != 2:
        raise generic.ScriptError("Availability years must be an array with exactly two values", value.pos)

    min_year = nmlop.SUB(value.values[0], 1920, value.pos)
    min_year = nmlop.MAX(min_year, 0)
    min_year = nmlop.MIN(min_year, 255)

    max_year = nmlop.SUB(value.values[1], 1920, value.pos)
    max_year = nmlop.MAX(max_year, 0)
    max_year = nmlop.MIN(max_year, 255)
    max_year = nmlop.SHIFT_LEFT(max_year, 8)

    return nmlop.OR(max_year, min_year).reduce()


def house_prop_21_22(value, index):
    # Take one of the values from the years_available array
    if not isinstance(value, Array) or len(value.values) != 2:
        raise generic.ScriptError("Availability years must be an array with exactly two values", value.pos)
    return value.values[index]


def house_random_colours(value):
    # User sets array with 4 values (range 0..15)
    # Output is a dword, each byte being a value from the array
    if not isinstance(value, Array) or len(value.values) != 4:
        raise generic.ScriptError("Random colours must be an array with exactly four values", value.pos)

    ret = None
    for i, colour in enumerate(value.values):
        if isinstance(colour, ConstantNumeric):
            generic.check_range(colour.value, 0, 15, "Random house colours", colour.pos)
        byte = nmlop.AND(colour, 0xFF)
        if i == 0:
            ret = byte
        else:
            byte = nmlop.SHIFT_LEFT(byte, i * 8)
            ret = nmlop.OR(ret, byte)
    return ret.reduce()


def house_available_mask(value):
    # User sets [town_zones, climates] array
    # Which is mapped to (town_zones | (climates & 0x800) | ((climates & 0xF) << 12))
    if not isinstance(value, Array) or len(value.values) != 2:
        raise generic.ScriptError("availability_mask must be an array with exactly 2 values", value.pos)

    climates = nmlop.AND(value.values[1], 0xF, value.pos)
    climates = nmlop.SHIFT_LEFT(climates, 12)
    above_snow = nmlop.AND(value.values[1], 0x800, value.pos)

    ret = nmlop.OR(climates, value.values[0])
    ret = nmlop.OR(ret, above_snow)
    return ret.reduce()


# List of valid IDs of old house types
old_houses = {
    0: set(),  # 1x1, see below
    2: {74, 76, 87},  # 2x1
    3: {7, 66, 68, 99},  # 1x2
    4: {20, 32, 40},  # 2x2
}
# All houses not part of a multitile-house, are 1x1 houses
old_houses[0] = set(range(110)).difference(
    house + i
    for house in (itertools.chain(*list(old_houses.values())))
    for i in range(4 if house in old_houses[4] else 2)
)


def mt_house_old_id(value, num_ids, size_bit):
    # For substitute / override properties
    # Set value for tile i (0 .. 3) to (value + i)
    # Also validate that the size of the old house matches
    if isinstance(value, ConstantNumeric) and value.value not in old_houses[size_bit]:
        raise generic.ScriptError(
            "Substitute / override house type must have the same size as the newly defined house.", value.pos
        )
    ret = [value]
    for i in range(1, num_ids):
        ret.append(nmlop.ADD(value, i).reduce())
    return ret


def mt_house_prop09(value, num_ids, size_bit):
    # Only bit 5 should be set for additional tiles
    # Additionally, correctly set the size bit (0, 2, 3 or 4) for the first tile
    if isinstance(value, ConstantNumeric) and (value.value & 0x1D) != 0:
        raise generic.ScriptError("Invalid bits set in house property 'building_flags'.", value.pos)
    ret = [nmlop.OR(value, 1 << size_bit).reduce()]
    for _i in range(1, num_ids):
        ret.append(nmlop.AND(value, 1 << 5).reduce())
    return ret


def mt_house_mask(mask, value, num_ids, size_bit):
    # Mask out the bits not present in the 'mask' parameter for additional tiles
    ret = [value]
    for _i in range(1, num_ids):
        ret.append(nmlop.AND(value, mask).reduce())
    return ret


def mt_house_zero(value, num_ids, size_bit):
    return [value] + (num_ids - 1) * [ConstantNumeric(0, value.pos)]


def mt_house_same(value, num_ids, size_bit):
    # Set to the same value for all tiles
    return num_ids * [value]


def mt_house_class(value, num_ids, size_bit):
    # Set class to 0xFF for additional tiles
    return [value] + (num_ids - 1) * [ConstantNumeric(0xFF, value.pos)]


# fmt: off
properties[0x07] = {
    "substitute":              {"size": 1, "num": 0x08, "multitile_function": mt_house_old_id, "first": None},
    "building_flags": two_byte_property(
        0x09,
        0x19,
        {"multitile_function": mt_house_prop09},
        {"multitile_function": lambda *args: mt_house_mask(0xFE, *args)}
    ),
    "years_available": [
        {"size": 2, "num": 0x0A, "multitile_function": mt_house_zero, "value_function": house_prop_0A},
        {
            "size": 2,
            "num": 0x21,
            "multitile_function": mt_house_zero,
            "value_function": lambda value: house_prop_21_22(value, 0),
        },
        {
            "size": 2,
            "num": 0x22,
            "multitile_function": mt_house_zero,
            "value_function": lambda value: house_prop_21_22(value, 1),
        },
    ],
    "population":              {"size": 1, "num": 0x0B, "multitile_function": mt_house_zero},
    "mail_multiplier":         {"size": 1, "num": 0x0C, "multitile_function": mt_house_zero},
    # prop 0D - 0F are replaced by prop 23
    "local_authority_impact":  {"size": 2, "num": 0x10, "multitile_function": mt_house_same},
    "removal_cost_multiplier": {"size": 1, "num": 0x11, "multitile_function": mt_house_same},
    "name":                    {"size": 2, "num": 0x12, "string": 0xDC, "multitile_function": mt_house_same},
    "availability_mask": {
        "size": 2,
        "num": 0x13,
        "multitile_function": mt_house_zero,
        "value_function": house_available_mask,
    },
    # prop 14 (callback flags 1) is not set by user
    "override":                {"size": 1, "num": 0x15, "multitile_function": mt_house_old_id},
    "refresh_multiplier":      {"size": 1, "num": 0x16, "multitile_function": mt_house_same},
    "random_colours": {
        "size": 4,
        "num": 0x17,
        "multitile_function": mt_house_same,
        "value_function": house_random_colours,
    },
    "probability":             {"size": 1, "num": 0x18, "multitile_function": mt_house_zero, "unit_conversion": 16},
    # prop 19 is the high byte of prop 09
    "animation_info": {
        "size": 1,
        "num": 0x1A,
        "multitile_function": mt_house_same,
        "value_function": lambda value: animation_info(value, 7, 128),
    },
    "animation_speed":         {"size": 1, "num": 0x1B, "multitile_function": mt_house_same},
    "building_class":          {"size": 1, "num": 0x1C, "multitile_function": mt_house_class},
    # prop 1D (callback flags 2) is not set by user
    # prop 1E is replaced by prop 23
    "minimum_lifetime":        {"size": 1, "num": 0x1F, "multitile_function": mt_house_zero},
    "watched_cargo_types": {
        "multitile_function": mt_house_same,
        "custom_function": lambda *values: ctt_list(0x20, *values),
    },
    # prop 21 -22 see above (years_available, prop 0A)
    "accepted_cargos": {
        "multitile_function": mt_house_same,
        "custom_function": lambda *values: accepted_cargos(0x23, *values),
    },
}
# fmt: on

# Feature 0x08 (General Vars) is implemented elsewhere (e.g. basecost, snowline)

#
# Feature 0x09 (Industry Tiles)
#

# fmt: off
properties[0x09] = {
    "substitute":          {"size": 1, "num": 0x08, "first": None},
    "override":            {"size": 1, "num": 0x09},
    # prop 0A - 0C are replaced by prop 13
    "land_shape_flags":    {"size": 1, "num": 0x0D},
    # prop 0E (callback flags) is not set by user
    "animation_info":      {"size": 2, "num": 0x0F, "value_function": animation_info},
    "animation_speed":     {"size": 1, "num": 0x10},
    "animation_triggers":  {"size": 1, "num": 0x11},
    "special_flags":       {"size": 1, "num": 0x12},
    "accepted_cargos":     {"custom_function": lambda value: accepted_cargos(0x13, value)},
}
# fmt: on

#
# Feature 0x0A (Industries)
#


class IndustryLayoutProp(BaseAction0Property):
    def __init__(self, layout_list):
        self.layout_list = layout_list

    def write(self, file):
        file.print_bytex(0x0A)
        file.print_byte(len(self.layout_list))
        # -6 because prop_num, num_layouts and size should not be included
        file.print_dword(self.get_size() - 6)
        file.newline()
        for layout in self.layout_list:
            layout.write(file)
        file.newline()

    def get_size(self):
        size = 6
        for layout in self.layout_list:
            size += layout.get_size()
        return size


def industry_layouts(value):
    if not isinstance(value, Array) or not all(isinstance(x, Identifier) for x in value.values):
        raise generic.ScriptError("layouts must be an array of layout names", value.pos)
    layouts = []
    for name in value.values:
        if name.value not in tilelayout_names:
            raise generic.ScriptError("Unknown layout name '{}'".format(name.value), name.pos)
        layouts.append(tilelayout_names[name.value])
    return [IndustryLayoutProp(layouts)]


def industry_prod_multiplier(value):
    if not isinstance(value, Array) or len(value.values) > 2:
        raise generic.ScriptError("Prod multiplier must be an array of up to two values", value.pos)
    props = []
    for i in range(0, 2):
        val = value.values[i].reduce_constant() if i < len(value.values) else ConstantNumeric(0)
        props.append(Action0Property(0x12 + i, val, 1))
    return props


class RandomSoundsProp(BaseAction0Property):
    def __init__(self, sound_list):
        self.sound_list = sound_list

    def write(self, file):
        file.print_bytex(0x15)
        file.print_byte(len(self.sound_list))
        for sound in self.sound_list:
            sound.write(file, 1)
        file.newline()

    def get_size(self):
        return len(self.sound_list) + 2


def random_sounds(value):
    if not isinstance(value, Array) or not all(isinstance(x, ConstantNumeric) for x in value.values):
        raise generic.ScriptError("random_sound_effects must be an array with sounds effects", value.pos)
    return [RandomSoundsProp(value.values)]


class ConflictingTypesProp(BaseAction0Property):
    def __init__(self, types_list):
        self.types_list = types_list
        assert len(self.types_list) == 3

    def write(self, file):
        file.print_bytex(0x16)
        for type in self.types_list:
            type.write(file, 1)
        file.newline()

    def get_size(self):
        return len(self.types_list) + 1


def industry_conflicting_types(value):
    if not isinstance(value, Array):
        raise generic.ScriptError("conflicting_ind_types must be an array of industry types", value.pos)
    if len(value.values) > 3:
        raise generic.ScriptError("conflicting_ind_types may have at most three entries", value.pos)

    types_list = []
    for val in value.values:
        types_list.append(val.reduce_constant())
    while len(types_list) < 3:
        types_list.append(ConstantNumeric(0xFF))
    return [ConflictingTypesProp(types_list)]


def industry_input_multiplier(value, prop_num):
    if not isinstance(value, Array) or len(value.values) > 2:
        raise generic.ScriptError("Input multiplier must be an array of up to two values", value.pos)
    val1 = value.values[0].reduce() if len(value.values) > 0 else ConstantNumeric(0)
    val2 = value.values[1].reduce() if len(value.values) > 1 else ConstantNumeric(0)
    if not isinstance(val1, (ConstantNumeric, ConstantFloat)) or not isinstance(val2, (ConstantNumeric, ConstantFloat)):
        raise generic.ScriptError("Expected a compile-time constant", value.pos)
    generic.check_range(val1.value, 0, 256, "input_multiplier", val1.pos)
    generic.check_range(val2.value, 0, 256, "input_multiplier", val2.pos)
    mul1 = int(val1.value * 256)
    mul2 = int(val2.value * 256)
    return [Action0Property(prop_num, ConstantNumeric(mul1 | (mul2 << 16)), 4)]


class IndustryInputMultiplierProp(BaseAction0Property):
    def __init__(self, prop_num, data):
        self.prop_num = prop_num
        self.data = data

    def write(self, file):
        file.print_bytex(self.prop_num)
        if len(self.data) == 0:
            file.print_byte(0)
            file.print_byte(0)
        else:
            file.print_byte(len(self.data))
            file.print_byte(len(self.data[0]))  # assume all sub-arrays are equal length
            file.newline()
            for out_muls in self.data:
                for mul in out_muls:
                    file.print_wordx(mul)
                file.newline()

    def get_size(self):
        if len(self.data) == 0:
            return 3
        else:
            return 3 + len(self.data) * len(self.data[0]) * 2


def industry_cargo_types(value):
    if isinstance(value, Array):
        cargo_types = value.values
    else:
        cargo_types = [value]
    if not all(isinstance(item, (ProduceCargo, AcceptCargo)) for item in value.values):
        raise generic.ScriptError(
            "Cargo types definition must be an array produce_cargo() and accept_cargo() expressions", value.pos
        )

    # collect all the cargo types involved
    input_cargos = []
    output_cargos = []

    def check_produce(prd):
        if len(prd.value) != 1:
            raise generic.ScriptError("Cargo types produce_cargo() expressions require 2 arguments", prd.pos)
        if not isinstance(prd.value[0], (ConstantNumeric, ConstantFloat)):
            raise generic.ScriptError(
                "Cargo types produce_cargo() expressions must have numeric constant values", prd.pos
            )
        if prd.cargotype not in output_cargos:
            output_cargos.append(prd.cargotype)

    def check_accept(acp):
        if item.cargotype not in input_cargos:
            input_cargos.append(item.cargotype)
        for outitem in item.value:
            if isinstance(outitem, ProduceCargo):
                check_produce(outitem)
            else:
                raise generic.ScriptError(
                    "Cargo types accept_cargo() expressions must only contain produce_cargo() expressions", outitem.pos
                )

    for item in cargo_types:
        # use "if not in: append" idiom rather than sets to preserve ordering of cargotypes between NML and NFO
        if isinstance(item, ProduceCargo):
            check_produce(item)
        elif isinstance(item, AcceptCargo):
            check_accept(item)
        else:
            raise AssertionError()

    if len(input_cargos) > 16:
        raise generic.ScriptError(
            "Cargo types definition contains more than 16 different accept_cargo() cargotypes", value.pos
        )
    if len(output_cargos) > 16:
        raise generic.ScriptError(
            "Cargo types definition contains more than 16 different produce_cargo() cargotypes", value.pos
        )

    # prepare lists for the remaining output properties
    prod_multipliers = [0] * len(output_cargos)
    input_multipliers = [[0] * len(output_cargos) for _ in input_cargos]
    has_inpmult = False

    # populate prod_multipliers and input_multipliers
    for item in cargo_types:
        if isinstance(item, ProduceCargo):
            prod_multipliers[output_cargos.index(item.cargotype)] = int(item.value[0].value)
        elif isinstance(item, AcceptCargo):
            row = input_multipliers[input_cargos.index(item.cargotype)]
            for outitem in item.value:
                row[output_cargos.index(outitem.cargotype)] = int(outitem.value[0].value * 256)
                if outitem.value[0].value > 0:
                    has_inpmult = True

    return [
        VariableByteListProp(0x25, [output_cargos]),
        VariableByteListProp(0x26, [input_cargos]),
        VariableByteListProp(0x27, [prod_multipliers]),
        IndustryInputMultiplierProp(0x28, input_multipliers if has_inpmult else []),
    ]


# fmt: off
properties[0x0A] = {
    "substitute":             {"size": 1, "num": 0x08, "first": None},
    "override":               {"size": 1, "num": 0x09},
    "layouts":                {"custom_function": industry_layouts},  # = prop 0A
    "life_type":              {"size": 1, "num": 0x0B},
    "closure_msg":            {"size": 2, "num": 0x0C, "string": 0xDC},
    "prod_increase_msg":      {"size": 2, "num": 0x0D, "string": 0xDC},
    "prod_decrease_msg":      {"size": 2, "num": 0x0E, "string": 0xDC},
    "fund_cost_multiplier":   {"size": 1, "num": 0x0F},
    "prod_cargo_types": {
        "size": 2,
        "num": 0x10,
        "value_function": lambda value: cargo_list(value, 2),
        "replaced_by": "cargo_types",
    },
    "accept_cargo_types": {
        "size": 4,
        "num": 0x11,
        "value_function": lambda value: cargo_list(value, 3),
        "replaced_by": "cargo_types",
    },
    # prop 12,13
    "prod_multiplier": {
        "custom_function": industry_prod_multiplier,
        "replaced_by": "cargo_types",
    },
    "min_cargo_distr":        {"size": 1, "num": 0x14},
    "random_sound_effects":   {"custom_function": random_sounds},  # = prop 15
    "conflicting_ind_types":  {"custom_function": industry_conflicting_types},  # = prop 16
    "prob_random":            {"size": 1, "num": 0x17},  # Obsolete, ambiguous name, use "prob_map_gen" instead
    "prob_map_gen":           {"size": 1, "num": 0x17},
    "prob_in_game":           {"size": 1, "num": 0x18},
    "map_colour":             {"size": 1, "num": 0x19},
    "spec_flags":             {"size": 4, "num": 0x1A},
    "new_ind_msg":            {"size": 2, "num": 0x1B, "string": 0xDC},
    "input_multiplier_1": {
        "custom_function": lambda value: industry_input_multiplier(value, 0x1C),
        "replaced_by": "cargo_types",
    },
    "input_multiplier_2": {
        "custom_function": lambda value: industry_input_multiplier(value, 0x1D),
        "replaced_by": "cargo_types",
    },
    "input_multiplier_3": {
        "custom_function": lambda value: industry_input_multiplier(value, 0x1E),
        "replaced_by": "cargo_types",
    },
    "name":                   {"size": 2, "num": 0x1F, "string": 0xDC},
    "prospect_chance":        {"size": 4, "num": 0x20, "unit_conversion": 0xFFFFFFFF},
    # prop 21, 22 (callback flags) are not set by user
    "remove_cost_multiplier": {"size": 4, "num": 0x23},
    "nearby_station_name":    {"size": 2, "num": 0x24, "string": 0xDC},
    # prop 25+26+27+28 combined in one structure
    "cargo_types":            {"custom_function": industry_cargo_types},
}
# fmt: on

#
# Feature 0x0B (Cargos)
#

# fmt: off
properties[0x0B] = {
    "number":                    {"num": 0x08, "size": 1},
    "type_name":                 {"num": 0x09, "size": 2, "string": 0xDC},
    "unit_name":                 {"num": 0x0A, "size": 2, "string": 0xDC},
    # Properties 0B, 0C are not used by OpenTTD
    "type_abbreviation":         {"num": 0x0D, "size": 2, "string": 0xDC},
    "sprite":                    {"num": 0x0E, "size": 2},
    "weight":                    {"num": 0x0F, "size": 1, "unit_type": "weight", "unit_conversion": 16},
    "penalty_lowerbound":        {"num": 0x10, "size": 1},
    "single_penalty_length":     {"num": 0x11, "size": 1},
    # 10 units of cargo across 20 tiles, with time factor = 255
    "price_factor":              {"num": 0x12, "size": 4, "unit_conversion": (1 << 21, 10 * 20 * 255)},
    "station_list_colour":       {"num": 0x13, "size": 1},
    "cargo_payment_list_colour": {"num": 0x14, "size": 1},
    "is_freight":                {"num": 0x15, "size": 1},
    "cargo_classes":             {"num": 0x16, "size": 2},
    "cargo_label":               {"num": 0x17, "size": 4, "string_literal": 4},
    "town_growth_effect":        {"num": 0x18, "size": 1},
    "town_growth_multiplier":    {"num": 0x19, "size": 2, "unit_conversion": 0x100},
    # 1A (callback flags) is not set by user
    "units_of_cargo":            {"num": 0x1B, "size": 2, "string": 0xDC},
    "items_of_cargo":            {"num": 0x1C, "size": 2, "string": 0xDC},
    "capacity_multiplier":       {"num": 0x1D, "size": 2, "unit_conversion": 0x100},
    "town_production_effect":    {"num": 0x1E, "size": 1},
    "town_production_multiplier": {"num": 0x1F, "size": 2, "unit_conversion": 0x100},
}
# fmt: on

# Feature 0x0C (Sound Effects) is implemented differently

#
# Feature 0x0D (Airports)
#


def airport_years(value):
    if not isinstance(value, Array) or len(value.values) != 2:
        raise generic.ScriptError("Availability years must be an array with exactly two values", value.pos)
    min_year = value.values[0].reduce_constant()
    max_year = value.values[1].reduce_constant()
    return [Action0Property(0x0C, ConstantNumeric(max_year.value << 16 | min_year.value), 4)]


class AirportLayoutProp(BaseAction0Property):
    def __init__(self, layout_list):
        self.layout_list = layout_list

    def write(self, file):
        file.print_bytex(0x0A)
        file.print_byte(len(self.layout_list))
        # -6 because prop_num, num_layouts and size should not be included
        file.print_dword(self.get_size() - 6)
        file.newline()
        for layout in self.layout_list:
            file.print_bytex(layout.properties["rotation"].value)
            layout.write(file)
        file.newline()

    def get_size(self):
        size = 6
        for layout in self.layout_list:
            size += layout.get_size() + 1
        return size


def airport_layouts(value):
    if not isinstance(value, Array) or not all(isinstance(x, Identifier) for x in value.values):
        raise generic.ScriptError("layouts must be an array of layout names", value.pos)
    layouts = []
    for name in value.values:
        if name.value not in tilelayout_names:
            raise generic.ScriptError("Unknown layout name '{}'".format(name.value), name.pos)
        layout = tilelayout_names[name.value]
        if "rotation" not in layout.properties:
            raise generic.ScriptError("Airport layouts must have the 'rotation' property", layout.pos)
        if layout.properties["rotation"].value not in (0, 2, 4, 6):
            raise generic.ScriptError(
                "Airport layout rotation is not a valid direction.", layout.properties["rotation"].pos
            )
        layouts.append(layout)
    return [AirportLayoutProp(layouts)]


# fmt: off
properties[0x0D] = {
    "override":         {"size": 1, "num": 0x08, "first": None},
    # 09 does not exist
    "layouts":          {"custom_function": airport_layouts},  # = prop 0A
    # 0B does not exist
    "years_available":  {"custom_function": airport_years},  # = prop 0C
    "ttd_airport_type": {"size": 1, "num": 0x0D},
    "catchment_area":   {"size": 1, "num": 0x0E},
    "noise_level":      {"size": 1, "num": 0x0F},
    "name":             {"size": 2, "num": 0x10, "string": 0xDC},
    "maintenance_cost": {"size": 2, "num": 0x11},
}
# fmt: on

# Feature 0x0E (Signals) doesn't currently have any action0

#
# Feature 0x0F (Objects)
#


def object_size(value):
    if not isinstance(value, Array) or len(value.values) != 2:
        raise generic.ScriptError("Object size must be an array with exactly two values", value.pos)
    sizex = value.values[0].reduce_constant()
    sizey = value.values[1].reduce_constant()
    if sizex.value < 1 or sizex.value > 15 or sizey.value < 1 or sizey.value > 15:
        raise generic.ScriptError("The size of an object must be at least 1x1 and at most 15x15 tiles", value.pos)
    return [Action0Property(0x0C, ConstantNumeric(sizey.value << 4 | sizex.value), 1)]


# fmt: off
properties[0x0F] = {
    "class":                  {"size": 4, "num": 0x08, "first": None, "string_literal": 4},
    # strings might be according to specs be either 0xD0 or 0xD4
    "classname":              {"size": 2, "num": 0x09, "string": 0xD0},
    "name":                   {"size": 2, "num": 0x0A, "string": 0xD0, "required": True},
    "climates_available":     {"size": 1, "num": 0x0B, "required": True},
    "size":                   {"custom_function": object_size, "required": True},  # = prop 0C
    "build_cost_multiplier":  {"size": 1, "num": 0x0D},
    "introduction_date":      {"size": 4, "num": 0x0E},
    "end_of_life_date":       {"size": 4, "num": 0x0F},
    "object_flags":           {"size": 2, "num": 0x10},
    "animation_info":         {"size": 2, "num": 0x11, "value_function": animation_info},
    "animation_speed":        {"size": 1, "num": 0x12},
    "animation_triggers":     {"size": 2, "num": 0x13},
    "remove_cost_multiplier": {"size": 1, "num": 0x14},
    # 15 (callback flags) is not set by user
    "height":                 {"size": 1, "num": 0x16},
    "num_views":              {"size": 1, "num": 0x17},
    "count_per_map256":       {"size": 1, "num": 0x18},
}
# fmt: on

#
# General tracktype properties that apply to features 0x10 & 0x12/13 (rail/road/tramtypes)
#


class LabelListProp(BaseAction0Property):
    def __init__(self, prop_num, labels):
        self.prop_num = prop_num
        self.labels = labels

    def write(self, file):
        file.print_bytex(self.prop_num)
        file.print_byte(len(self.labels))
        for label in self.labels:
            parse_string_to_dword(label)  # Error if the wrong length or not ASCII
            label.write(file, 4)
        file.newline()

    def get_size(self):
        return len(self.labels) * 4 + 2


def label_list(value, prop_num, description):
    if not isinstance(value, Array):
        raise generic.ScriptError(description + " list must be an array of literal strings", value.pos)
    return [LabelListProp(prop_num, value.values)]


# fmt: off
common_tracktype_props = {
    # label allocated during reservation stage, so doesn't need to be set first.
    "label":                {"size": 4, "num": 0x08, "string_literal": 4},
    "toolbar_caption":      {"size": 2, "num": 0x09, "string": 0xDC},
    "menu_text":            {"size": 2, "num": 0x0A, "string": 0xDC},
    "build_window_caption": {"size": 2, "num": 0x0B, "string": 0xDC},
    "autoreplace_text":     {"size": 2, "num": 0x0C, "string": 0xDC},
    "new_engine_text":      {"size": 2, "num": 0x0D, "string": 0xDC},
    "construction_cost":    {"size": 2, "num": 0x13},
    "map_colour":           {"size": 1, "num": 0x16},
    "introduction_date":    {"size": 4, "num": 0x17},
    "sort_order":           {"size": 1, "num": 0x1A},
    "name":                 {"size": 2, "num": 0x1B, "string": 0xDC},
    "maintenance_cost":     {"size": 2, "num": 0x1C},
}

#
# Feature 0x10 (Rail Types)
#

properties[0x10] = {
    **common_tracktype_props,
    "compatible_railtype_list":  {"custom_function": lambda x: label_list(x, 0x0E, "Railtype")},
    "powered_railtype_list":     {"custom_function": lambda x: label_list(x, 0x0F, "Railtype")},
    "railtype_flags":            {"size": 1, "num": 0x10},
    "curve_speed_multiplier":    {"size": 1, "num": 0x11},
    "station_graphics":          {"size": 1, "num": 0x12},
    "speed_limit": {
        "size": 2,
        "num": 0x14,
        "unit_type": "speed",
        "unit_conversion": (5000, 1397),
        "adjust_value": lambda val, unit: ottd_display_speed(val, 1, 1, unit),
    },
    "acceleration_model":        {"size": 1, "num": 0x15},
    "requires_railtype_list":    {"custom_function": lambda x: label_list(x, 0x18, "Railtype")},
    "introduces_railtype_list":  {"custom_function": lambda x: label_list(x, 0x19, "Railtype")},
    "alternative_railtype_list": {"custom_function": lambda x: label_list(x, 0x1D, "Railtype")},
}

#
# Feature 0x11 (Airport Tiles)
#

properties[0x11] = {
    "substitute":         {"size": 1, "num": 0x08, "first": None},
    "override":           {"size": 1, "num": 0x09},
    # 0A - 0D don't exist (yet?)
    # 0E (callback flags) is not set by user
    "animation_info":     {"size": 2, "num": 0x0F, "value_function": animation_info},
    "animation_speed":    {"size": 1, "num": 0x10},
    "animation_triggers": {"size": 1, "num": 0x11},
}

#
# Feature 0x12 (Road Types)
#

properties[0x12] = {
    **common_tracktype_props,
    "powered_roadtype_list":     {"custom_function": lambda x: label_list(x, 0x0F, "Roadtype")},
    "roadtype_flags":            {"size": 1, "num": 0x10},
    "speed_limit": {
        "size": 2,
        "num": 0x14,
        "unit_type": "speed",
        "unit_conversion": (10000, 1397),
        "adjust_value": lambda val, unit: ottd_display_speed(val, 1, 2, unit)
    },
    "requires_roadtype_list":    {"custom_function": lambda x: label_list(x, 0x18, "Roadtype")},
    "introduces_roadtype_list":  {"custom_function": lambda x: label_list(x, 0x19, "Roadtype")},
    "alternative_roadtype_list": {"custom_function": lambda x: label_list(x, 0x1D, "Roadtype")},
}

#
# Feature 0x13 (Tram Types)
#

properties[0x13] = {
    **common_tracktype_props,
    "powered_tramtype_list":     {"custom_function": lambda x: label_list(x, 0x0F, "Tramtype")},
    "tramtype_flags":            {"size": 1, "num": 0x10},
    "speed_limit": {
        "size": 2,
        "num": 0x14,
        "unit_type": "speed",
        "unit_conversion": (10000, 1397),
        "adjust_value": lambda val, unit: ottd_display_speed(val, 1, 2, unit),
    },
    "requires_tramtype_list":    {"custom_function": lambda x: label_list(x, 0x18, "Tramtype")},
    "introduces_tramtype_list":  {"custom_function": lambda x: label_list(x, 0x19, "Tramtype")},
    "alternative_tramtype_list": {"custom_function": lambda x: label_list(x, 0x1D, "Tramtype")},
}

#
# Feature 0x14 (Road stops)
#


class ByteSequenceProp(BaseAction0Property):
    def __init__(self, prop_num, items, description, pos):
        self.prop_num = prop_num
        self.items = items
        self.description = description
        self.pos = pos

    def write(self, file):
        file.print_bytex(self.prop_num)
        for item in self.items:
            val = item.reduce_constant().value
            if val > 0xFF or val < 0:
                raise generic.ScriptError(self.description + " items must be bytes", self.pos)
            file.print_byte(val)
        file.newline()

    def get_size(self):
        return 1 + len(self.items)


def byte_sequence_list(value, prop_num, description, expected_count):
    if not isinstance(value, Array) or len(value.values) != expected_count:
        raise generic.ScriptError(description + " must be an array of " + str(expected_count) + " bytes", value.pos)
    return [ByteSequenceProp(prop_num, value.values, description, value.pos)]


properties[0x14] = {
    'class':                     {'size': 4, 'num': 0x08, "first": None, "string_literal": 4},
    'availability_type':         {'size': 1, 'num': 0x09},
    'name':                      {'size': 2, 'num': 0x0A, "string": 0xDC, "required": True},
    'classname':                 {'size': 2, 'num': 0x0B, "string": 0xDC},
    'draw_mode':                 {'size': 1, 'num': 0x0C},
    "cargo_random_triggers":     {"size": 4, "num": 0x0D},
    "animation_info":            {"size": 2, "num": 0x0E, "value_function": animation_info},
    "animation_speed":           {"size": 1, "num": 0x0F},
    "animation_triggers":        {"size": 2, "num": 0x10},
    # 11 (callback flags) is not set by user
    "general_flags":             {"size": 4, "num": 0x12},
    "heights":                   {"custom_function": lambda x: station_tile_list(x, 0x13, "Station height")},
    "blocked_pillars":           {"custom_function": lambda x: station_tile_list(x, 0x14, "Station blocked pillar")},
    "cost_multipliers":          {"custom_function": lambda x: byte_sequence_list(x, 0x15, "Cost multipliers", 2)},
}



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, global_constants
from nml.actions import action7
from nml.ast import base_statement


class ConditionalList(base_statement.BaseStatementList):
    """
    Wrapper for a complete if/else if/else if/else block.
    """

    def __init__(self, conditionals):
        assert len(conditionals) > 0
        base_statement.BaseStatementList.__init__(
            self,
            "if/else-block",
            conditionals[0].pos,
            base_statement.BaseStatementList.LIST_TYPE_SKIP,
            conditionals,
            in_item=True,
        )

    def get_action_list(self):
        return action7.parse_conditional_block(self)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Conditional")
        base_statement.BaseStatementList.debug_print(self, indentation + 2)

    def __str__(self):
        ret = ""
        ret += " else ".join([str(stmt) for stmt in self.statements])
        ret += "\n"
        return ret


class Conditional(base_statement.BaseStatementList):
    """
    Condition along with the code that has to be executed if the condition
    evaluates to some value not equal to 0.

    @ivar expr: The expression where the execution of code in this block depends on.
    @type expr: L{Expression}
    """

    def __init__(self, expr, block, pos):
        base_statement.BaseStatementList.__init__(
            self, "if/else-block", pos, base_statement.BaseStatementList.LIST_TYPE_SKIP, block, in_item=True
        )
        self.expr = expr

    def pre_process(self):
        if self.expr is not None:
            self.expr = self.expr.reduce(global_constants.const_list)
        base_statement.BaseStatementList.pre_process(self)

    def debug_print(self, indentation):
        if self.expr is not None:
            generic.print_dbg(indentation, "Expression:")
            self.expr.debug_print(indentation + 2)

        generic.print_dbg(indentation, "Block:")
        base_statement.BaseStatementList.debug_print(self, indentation + 2)

    def __str__(self):
        ret = ""
        if self.expr is not None:
            ret += "if ({})".format(self.expr)
        ret += " {\n"
        ret += base_statement.BaseStatementList.__str__(self)
        ret += "}\n"
        return ret
